# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for invokable resources."""

import contextlib
import dataclasses
import inspect
import traceback as traceback_module
from typing import (
  Any, Callable, Generic, Iterable, List, Mapping, Optional, Tuple, Type,
  TypeVar, cast)

from enact import contexts
from enact import interfaces
from enact import references
from enact import resources
from enact import resource_registry


AnyT = TypeVar('AnyT')
ResourceT = TypeVar('ResourceT', bound=interfaces.ResourceBase)
ExceptionT = TypeVar('ExceptionT', bound='ExceptionResource')


@resource_registry.register
class ExceptionResource(interfaces.ResourceBase, Exception):
  """A resource that is also an exception."""

  def __init__(self, *args):
    Exception.__init__(self, *args)

  @classmethod
  def field_names(cls) -> Iterable[str]:
    """Returns the names of the fields of the resource."""
    yield 'args'

  def field_values(self) -> Iterable[interfaces.FieldValue]:
    """Return a list of field values, aligned with field_names."""
    yield resource_registry.to_field_value(self.args)

  @classmethod
  def from_fields(cls: Type[ResourceT],
                  field_dict: Mapping[str, interfaces.FieldValue]) -> ResourceT:
    """Constructs the resource from a value dictionary."""
    return cls(*resource_registry.from_field_value(field_dict['args']))

  def set_from(self: ResourceT, other: Any):
    """Sets the fields of this resource from another resource."""
    super().set_from(other)  # Raise error


@resource_registry.register
class NativeException(ExceptionResource):
  """A python exception wrapped as a resource."""

  def __init__(
      self, type_name: str, str_representation: str, traceback: str):
    super().__init__(type_name, str_representation, traceback)

  @property
  def type_name(self) -> str:
    return self.args[0]

  @property
  def str_representation(self) -> str:
    return self.args[1]

  @property
  def traceback(self) -> str:
    return self.args[2]

  def __str__(self) -> str:
    return f'{self.type_name}: {self.str_representation}'


# Input value type.
I_contra = TypeVar('I_contra', contravariant=True)
# Output value type.
O_co = TypeVar('O_co', covariant=True)


@resource_registry.register
class InputRequest(ExceptionResource):
  """An exception indicating that external input is required."""

  def __init__(
      self,
      invokable: references.Ref[Callable],
      for_value: references.Ref,
      requested_output: Type,
      context: Any):
    """Initialize a new input request.

    Args:
      invokable: The parent invokable that issued the request.
      for_input: A reference to the value input is requested for.
      requested_output: The type of output that is requested.
      context: Any additional context required to resolve the input request.
    """
    if not references.Store.get_current():
      raise contexts.NoActiveContext(
        'InputRequest must be created within a Store context.')
    super().__init__(
      invokable,
      for_value,
      requested_output,
      context)

  @property
  def invokable(self) -> references.Ref[Callable]:
    """Returns the invokable that requested the input."""
    return self.args[0]

  @property
  def for_value(self) -> references.Ref:
    """Returns a reference to the resource for which input is requested."""
    return self.args[1]

  @property
  def requested_type(self) -> Type:
    """Returns the type of input requested."""
    return self.args[2]

  @property
  def context(self) -> Any:
    return self.args[3]

  def continue_invocation(
      self,
      invocation: 'Invocation[I_contra, O_co]',
      value: Any,
      strict: bool=True) -> (
        'Invocation[I_contra, O_co]'):
    """Replays the invocation with the given value."""
    ref = references.commit(self)
    def _exception_override(exception_ref: references.Ref[ExceptionResource]):
      if exception_ref == ref:
        return references.commit(value)
      return None
    return invocation.replay(
      exception_override=_exception_override, strict=strict)

  async def continue_invocation_async(
      self,
      invocation: 'Invocation[I_contra, O_co]',
      value: Any,
      strict: bool=True):
    ref = references.commit(self)
    def _exception_override(exception_ref: references.Ref[ExceptionResource]):
      if exception_ref == ref:
        return references.commit(value)
      return None
    return await invocation.replay_async(
      exception_override=_exception_override, strict=strict)


@resource_registry.register
class InputRequestOutsideInvocation(ExceptionResource):
  """Raised when input required is called outside an invocation."""


@resource_registry.register
class InvocationError(ExceptionResource):
  """An error during an invocation."""


@resource_registry.register
class InvokableTypeError(InvocationError, TypeError):
  """An type error on a callable input or output."""


@resource_registry.register
class InputChanged(InvocationError):
  """Raised when an input changes during invocation."""


@resource_registry.register
class RequestedTypeUndetermined(InvocationError):
  """Raised when the requested type cannot be determined."""


def _request_input(
    for_input: Any,
    requested_type: Optional[Type],
    context: Any=None):
  """Requests an input from a user or external system.

  Args:
    for_resource: The resource for which input is requested.
    requested_type: The type of input requested. If not specified, the type will
      be inferred to be the output type of the current invocation.
    context: Anything that provides context for the request.
  Raises:
    InputRequest: The input request exception.
    InputRequestOutsideInvocation: If the request was made outside an
      invocation.
  """
  builder: Optional[Builder] = Builder.get_current()
  if not builder:
    raise InputRequestOutsideInvocation(context, requested_type)
  requested_type = requested_type or builder.invokable.get_output_type()
  if not requested_type:
    raise RequestedTypeUndetermined(
      'Requested type must be specified when output type is undetermined.')
  raise InputRequest(
    references.commit(builder.invokable),
    references.commit(for_input),
    requested_type,
    context)


@resource_registry.register
@dataclasses.dataclass
class Request(Generic[I_contra, O_co], resources.Resource):
  """An invocation request."""
  invokable: references.Ref['_InvokableBase[I_contra, O_co]']
  input: references.Ref[I_contra]


@resource_registry.register
@dataclasses.dataclass
class Response(Generic[I_contra, O_co], resources.Resource):
  """An invocation response."""
  invokable: references.Ref[Callable]
  output: Optional[references.Ref[O_co]]
  # Exception raised during call.
  raised: Optional[references.Ref[ExceptionResource]]
  # Whether the exception was raised locally or propagated from a child.
  raised_here: bool
  # Subinvocations associated with this invocation.
  children: List[references.Ref['Invocation']]

  def is_complete(self) -> bool:
    """Returns whether this invocation is complete."""
    return self.raised is not None or self.output is not None


# A function that may override some exceptions that occur during invocation.
ExceptionOverride = Callable[[references.Ref[ExceptionResource]],
                             Optional[references.Ref]]



@resource_registry.register
@dataclasses.dataclass
class Invocation(Generic[I_contra, O_co], resources.Resource):
  """An invocation."""
  request: references.Ref[Request[I_contra, O_co]]
  response: references.Ref[Response[I_contra, O_co]]

  def successful(self) -> bool:
    """Returns true if the invocation completed successfully."""
    if not self.response:
      return False
    return self.response().output is not None

  def get_input(self) -> Any:
    """Returns the input."""
    return self.request().input()

  def get_output(self) -> O_co:
    """Returns the output or raise assertion error."""
    output = self.response().output
    assert output
    return output()

  def get_raised(self) -> ExceptionResource:
    """Returns the raised exception or raise assertion error."""
    raised = self.response().raised
    assert raised
    return raised()

  def get_raised_here(self) -> bool:
    """Whether the exception was originally raised here or in a child."""
    response = self.response()
    assert response.raised, 'No exception was raised.'
    return response.raised_here

  def get_children(self) -> Iterable['Invocation']:
    """Yields the child invocations."""
    children = self.response().children
    for child in children:
      yield child()

  def get_child(self, index: int) -> 'Invocation':
    """Returns the child invocation corresponding to the index."""
    children = self.response().children
    return children[index]()

  def clear_output(self):
    """Clear the output of the invocation."""
    with self.response.modify() as response:
      response.output = None

  def rewind(self, num_calls=1) -> 'Invocation[I_contra, O_co]':
    """Rewinds the invocation by the specified number of calls."""
    invocation = resource_registry.deepcopy(self)
    with invocation.response.modify() as response:
      response.output = None
      for _ in range(num_calls):
        if response.children:
          response.children.pop(-1)
    return invocation

  def replay(
      self,
      exception_override: ExceptionOverride=lambda x: None,
      strict: bool=True) -> (
        'Invocation[I_contra, O_co]'):
    """Replay the invocation, retrying exceptions or overiding them."""
    invokable = resource_registry.wrap(self.request().invokable())
    if isinstance(invokable, AsyncInvokableBase):
      raise InvocationError(
        'Cannot replay async invocations synchronously. '
        'Use the "replay_async" coroutine instead.')
    assert isinstance(invokable, InvokableBase)
    return invokable.invoke(
      self.request().input,
      replay_from=self,
      exception_override=exception_override,
      strict=strict)

  async def replay_async(
      self,
      exception_override: ExceptionOverride=lambda x: None,
      strict: bool=True) -> (
        'Invocation[I_contra, O_co]'):
    """Replay the invocation, retrying exceptions or overiding them."""
    invokable = resource_registry.wrap(self.request().invokable())
    if isinstance(invokable, InvokableBase):
      raise InvocationError(
        'Cannot replay synchronous invocations asynchronously. '
        'Use the "replay" function instead.')
    assert isinstance(invokable, AsyncInvokableBase)
    return await invokable.invoke(
      self.request().input,
      replay_from=self,
      exception_override=exception_override,
      strict=strict)


class ReplayError(InvocationError):
  """An error during replay."""


@contexts.register
class ReplayContext(Generic[I_contra, O_co], contexts.Context):
  """A replay of an invocation."""

  def __init__(
      self,
      subinvocations: Iterable[references.Ref[Invocation]],
      exception_override: ExceptionOverride=lambda x: None,
      strict: bool = True):
    """Create a new replay context.

    Args:
      subinvocations: The subinvocations to replay.
      exception_override: A function that may override exceptions that occur
        during replay. If an exception is not overriden, the invokable will
        be retried.
      strict: If true, an error will be raised when attempting to replay a
        subinvocation, but the provided subinvocation does not match the
        invokable and input. If false, a non-matching sub-invocation will
        be ignored and the corresponding invokable will be retried on its
        actual input.
    """
    super().__init__()
    self._exception_override = exception_override
    self._available_children = list(subinvocations)
    if not all(isinstance(x, references.Ref) for x in self._available_children):
      assert False
    self._strict = strict

  @classmethod
  def call_or_replay(
      cls,
      invokable: 'InvokableBase[I_contra, O_co]',
      arg: I_contra) -> O_co:
    """If there is a replay active, try to use it."""
    context: Optional[ReplayContext[I_contra, O_co]] = (
      ReplayContext.get_current())
    call = invokable.call

    if arg is None and len(inspect.signature(invokable.call).parameters) == 0:
      # Allow invokables that take no call args if they accept NoneResources.
      # pylint: disable=unnecessary-lambda-assignment
      call = lambda _: invokable.call()  # type: ignore

    if context:
      # pylint: disable=protected-access
      replayed_output, child_ctx = context._consume_replay(invokable, arg)
      if replayed_output is not None:
        return replayed_output()
      else:
        with child_ctx:
          return call(arg)
    return call(arg)

  @classmethod
  async def async_call_or_replay(
      cls,
      invokable: 'AsyncInvokableBase[I_contra, O_co]',
      arg: I_contra) -> O_co:
    """If there is a replay active, try to use it."""
    context: Optional[ReplayContext[I_contra, O_co]] = (
      ReplayContext.get_current())
    call = invokable.call
    if (arg is None and
        len(inspect.signature(invokable.call).parameters) == 0):
      # Allow invokables that take no call args if they accept NoneResources.
      # pylint: disable=unnecessary-lambda-assignment
      call = lambda _: invokable.call()  # type: ignore

    if context:
      # pylint: disable=protected-access
      replayed_output, child_ctx = context._consume_replay(invokable, arg)
      if replayed_output is not None:
        return replayed_output()
      else:
        with child_ctx:
          result = await call(arg)
          return result
    result = await call(arg)
    return result

  def _consume_replay(
      self,
      invokable: '_InvokableBase[I_contra, O_co]',
      input_resource: I_contra) -> Tuple[Optional[references.Ref[O_co]],
                                        'ReplayContext[I_contra, O_co]']:
    """Replay the invocation if possible and return a child context."""
    request = references.commit(Request(
      references.commit(invokable),
      references.commit(input_resource)))
    for i, child in enumerate(self._available_children):
      if child().request == request:
        break
      elif self._strict:
        raise ReplayError(
          f'Expected invocation {invokable}({input_resource}) but got '
          f'{child().request().invokable()}({child().request().input()}).\n'
          f'Ensure that calls to subinvokables are deterministic '
          f'or use strict=False.')
    else:
      # No matching replay found.
      return None, ReplayContext([], self._exception_override)

    # Consume child invocation
    self._available_children.pop(i)
    replay_response = child().response()
    replay_children = replay_response.children

    # Replay successful executions.
    if replay_response.output:
      invokable.set_from(resource_registry.wrap(replay_response.invokable()))
      Builder.register_replayed_subinvocations(replay_children)
      return replay_response.output, ReplayContext(
        replay_children,
        self._exception_override,
        self._strict)

    # Check for exception override
    if replay_response.raised and replay_response.raised_here:
      # We only override exceptions at the point of the stack where they were
      # originally raised.
      override = self._exception_override(replay_response.raised)
      invokable.set_from(resource_registry.wrap(replay_response.invokable()))
      if override is not None:
        # Typecheck the override.
        output_type = invokable.get_output_type()
        if output_type and not isinstance(override(), output_type):
          raise InvokableTypeError(
            f'Exception override {override()} is not of required type '
            f'{invokable.get_input_type()}.')
        Builder.register_replayed_subinvocations(replay_children)
        # Set invokable from response.
        return (
          cast(references.Ref[O_co], override),
          ReplayContext(replay_children,
                        self._exception_override,
                        self._strict))

    # Trigger reexecution of the invocation.
    return None, ReplayContext(
      replay_children,
      self._exception_override, self._strict)


class IncompleteSubinvocationError(InvocationError):
  """A subinvocation hasn't completed during the call."""


@contexts.register
class Builder(Generic[I_contra, O_co], contexts.Context):
  """A builder for invocations."""

  def __init__(
      self,
      invokable: '_InvokableBase[I_contra, O_co]',
      input_resource: references.Ref[I_contra]):
    """Initializes the builder."""
    super().__init__()
    self.invokable = invokable
    self.input_ref = input_resource

    self._children: Optional[List[Builder]] = None
    self._replayed_subinvocations: Optional[
      List[references.Ref[Invocation]]] = None
    self._request = Request(
      references.commit(invokable), input_resource)

    self._invocation: Optional[Invocation] = None
    self._parent: Optional[Builder] = self.get_current()
    if self._parent:
      self._parent.register_child(self)
    self.exception_raised: Optional[Exception] = None
    self.exception_wrapped: bool = False

  @property
  def completed(self) -> bool:
    """Returns true if the invocation is complete."""
    return self._invocation is not None

  def register_child(self, child: 'Builder'):
    """Registers a subinvocation."""
    if self._children is None:
      self._children = []
    self._children.append(child)

  @classmethod
  def register_replayed_subinvocations(
      cls, subinvocations: Iterable[references.Ref[Invocation]]):
    """Registers a list of subinvocations that were replayed."""
    builder = cls.get_current()
    if not builder:
      return
    # pylint: disable=protected-access
    builder._replayed_subinvocations = list(subinvocations)

  def _get_subinvocations(self) -> List[references.Ref[Invocation]]:
    """Returns the list of subinvocations."""
    assert (
      self._replayed_subinvocations is None or self._children is None), (
      'Cannot have both replayed and non-replayed subinvocations.')
    if self._replayed_subinvocations is not None:
      return self._replayed_subinvocations
    children = self._children or []
    for i, child in enumerate(children or []):
      if not child.completed:
        raise IncompleteSubinvocationError(
          f'Subinvocation {i} did not complete during invocation of parent:'
          f' {child.invokable} invoked on'
          f' {child.input_ref()}')
    return [references.commit(c.invocation) for c in children]

  def _is_child_exception(self, exception: Exception) -> bool:
    """Checks if the exception was originally raised by an immediate child."""
    children = self._children or []
    assert not self._replayed_subinvocations, (
      'Subinvocations were replayed, but an exception was raised.')
    return any(s.exception_raised is exception for s in children)

  @property
  def invocation(self) -> Invocation[I_contra, O_co]:
    assert self._invocation, (
      'The "call" function must be called before accessing the invocation.')
    return self._invocation

  def _check_call_valid(
      self,
      input_value: Any):
    """Checks that the call did not do invalid things."""
    if references.commit(input_value) != self.input_ref:
      raise InputChanged(
        f'Input changed during invocation of {self.invokable} on input '
        f'{input_value}. Only the invokable may change.')

  def _process_output(
      self,
      output: Any) -> references.Ref[O_co]:
    """Process the output and check for errors."""
    return cast(references.Ref[O_co], references.commit(output))

  def _wrap_exception(self, exception: Exception) -> ExceptionResource:
    """Wraps an exception if necessary."""
    if not isinstance(exception, ExceptionResource):
      self.exception_wrapped = True
      return NativeException(
        type_name=type(exception).__name__,
        str_representation=str(exception),
        traceback=traceback_module.format_exc())
    return exception

  def _create_invocation(
      self,
      output_ref: Optional[references.Ref[O_co]],
      exception: Optional[Exception]):
    """Process a call result and set the invocation object."""
    exception_ref: Optional[references.Ref[ExceptionResource]] = None
    raised_here = False
    if exception:
      exception_ref = references.commit(
        self._wrap_exception(exception))
      self.exception_raised = exception
      raised_here = not self._is_child_exception(exception)
    subinvocations = self._get_subinvocations()
    response: Response = Response(
      references.commit(self.invokable), output_ref,
      exception_ref, raised_here, children=subinvocations)
    self._invocation = Invocation(
      references.commit(self._request),
      references.commit(response))

  def call(self) -> O_co:
    """Call the invokable and set self.invocation."""
    with self:
      exception: Optional[Exception] = None
      output_ref: Optional[references.Ref[O_co]] = None
      try:
        input_value = self.input_ref()
        invokable = self.invokable
        assert isinstance(invokable, InvokableBase)
        output_value = ReplayContext.call_or_replay(
          invokable, input_value)
        self._check_call_valid(input_value)
        output_ref = self._process_output(output_value)
      except Exception as e:
        exception = e
        raise
      finally:
        self._create_invocation(output_ref, exception)
      return cast(O_co, output_value)

  async def async_call(self) -> O_co:
    """Call the async invokable and set self.invocation."""
    with self:
      exception: Optional[Exception] = None
      output_ref: Optional[references.Ref[O_co]] = None
      try:
        input_value = self.input_ref()
        invokable = self.invokable
        assert isinstance(invokable, AsyncInvokableBase)
        output_value = await ReplayContext.async_call_or_replay(
          invokable, input_value)
        self._check_call_valid(input_value)
        output_ref = self._process_output(output_value)
      except Exception as e:
        exception = e
        raise
      finally:
        self._create_invocation(output_ref, exception)
      return cast(O_co, output_value)

class _InvokableBase(Generic[I_contra, O_co], interfaces.ResourceBase):
  """Base class for sync / async invokable resources."""
  _enact_input_type: Optional[Type[I_contra]] = None
  _enact_output_type: Optional[Type[O_co]] = None

  @classmethod
  def get_input_type(cls) -> Optional[Type[I_contra]]:
    """Returns the type of the input if known."""
    return cls._enact_input_type

  def __call__(self, *args, **kwargs):
    """Subclasses implement this directly or as async."""
    raise NotImplementedError()

  def call(self, *args, **kwargs):
    """Subclasses implement this directly or as async."""
    raise NotImplementedError()

  @classmethod
  def _check_input_type(cls, value: Any):
    """Check the input type."""
    input_type: Optional[Type] = cls._enact_input_type
    # pylint: disable=isinstance-second-argument-not-valid-type
    if input_type and not isinstance(value, input_type):
      raise InvokableTypeError(
        f'Input must be of type {cls._enact_input_type}, '
        f'but got {type(value)}.')

  @classmethod
  def _check_output_type(cls, value: Any):
    """Check the output type."""
    output_type: Optional[Type] = cls._enact_output_type
    # pylint: disable=isinstance-second-argument-not-valid-type
    if output_type is not None and not isinstance(value, output_type):
      raise InvokableTypeError(
        f'Output must be of type {cls._enact_output_type}, '
        f'but got {type(value)}.')

  @classmethod
  def get_output_type(cls) -> Optional[Type[O_co]]:
    """Returns the type of the output if known."""
    return cls._enact_output_type

  @staticmethod
  def _process_invoke_arg(
      arg: Optional[references.Ref[I_contra]]) -> (
        references.Ref[I_contra]):
    """Processes an invocation argument."""
    if arg is None:
      arg = cast(references.Ref[I_contra],
                 references.commit(None))
    if not isinstance(arg, references.Ref):
      raise InvokableTypeError('Input must be a reference.')
    return arg

  @staticmethod
  def _invoke_exit_stack(
      replay_from: Optional[Invocation[I_contra, O_co]]=None,
      exception_override: ExceptionOverride=lambda _: None,
      strict: bool=False) -> contextlib.ExitStack:
    """Creates an exit stack for invoking an invokable."""
    exit_stack = contextlib.ExitStack()
    # Execute in a top-level context to ensure that there are no parents.
    exit_stack.enter_context(ReplayContext.top_level())
    exit_stack.enter_context(Builder.top_level())
    if replay_from:
      exit_stack.enter_context(ReplayContext(
        [references.commit(replay_from)],
        exception_override, strict))
    return exit_stack


class InvokableBase(_InvokableBase[I_contra, O_co]):
  """Base class for invokable resources."""

  def call(self, value: I_contra) -> O_co:
    """Executes the invokable."""
    raise NotImplementedError()

  def __call__(self, arg=None) -> O_co:
    """Executes the invokable, tracking invocation metadata."""
    parent: Optional[Builder] = Builder.get_current()

    self._check_input_type(arg)
    # Execution not tracked, so just call or replay the invokable.
    if not parent:
      output = ReplayContext.call_or_replay(self, arg)
    else:
      builder = Builder(self, references.commit(arg))
      output = builder.call()
    self._check_output_type(output)
    return output

  def invoke(
      self,
      arg: Optional[references.Ref[I_contra]]=None,
      replay_from: Optional[Invocation[I_contra, O_co]]=None,
      exception_override: ExceptionOverride=lambda _: None,
      raise_on_errors: Tuple[Type[Exception], ...]=(
        InvocationError, interfaces.FrameworkError),
      wrap_exceptions: bool=False,
      strict: bool=True,
      commit: bool=True) -> Invocation[I_contra, O_co]:
    """Invoke the invokable, tracking invocation metadata.

    Args:
      arg: The input resource.
      replay_from: An optional invocation to replay form.
      exception_override: If replaying, an optional override for replayed
        exceptions.
      raise_on_errors: Which errors should raise beyond the invocation.
      wrap_exceptions: Whether to wrap non-resource exceptions in
        NativeException. If False, non-resource exceptions raised during
        invocation will be re-raised.
      strict: Whether replay should fail if the replayed invocation
        does not match the current invocation.
      commit: Whether to commit the new invocation object.
    Returns:
      The invocation generated.
    """
    arg = self._process_invoke_arg(arg)

    with self._invoke_exit_stack(replay_from, exception_override, strict):
      builder = Builder(self, arg)
      try:
        builder.call()
      except raise_on_errors:
        raise
      except Exception:  # pylint: disable=broad-exception-caught
        should_raise_wrapped_error = (
          builder.exception_wrapped and not wrap_exceptions)
        if not builder.completed or should_raise_wrapped_error:
          raise
      invocation = builder.invocation
    if commit:
      references.commit(invocation)
    return invocation


class AsyncInvokableBase(_InvokableBase[I_contra, O_co]):
  """Base class for invokable resources."""

  async def call(self, value: I_contra) -> O_co:  # pylint: disable=invalid-overridden-method
    """Executes the async invokable."""
    raise NotImplementedError()

  async def __call__(self, arg=None) -> O_co:  # pylint: disable=invalid-overridden-method
    """Executes the async invokable, tracking invocation metadata."""
    parent: Optional[Builder] = Builder.get_current()

    self._check_input_type(arg)

    # Execution not tracked, so just call or replay the invokable.
    if not parent:
      output = await ReplayContext.async_call_or_replay(self, arg)
    else:
      builder = Builder(self, references.commit(arg))
      output = await builder.async_call()

    self._check_output_type(output)
    return output

  async def invoke(
      self,
      arg: Optional[references.Ref[I_contra]]=None,
      replay_from: Optional[Invocation[I_contra, O_co]]=None,
      exception_override: ExceptionOverride=lambda _: None,
      raise_on_errors: Tuple[Type[Exception], ...]=(
        InvocationError, interfaces.FrameworkError),
      wrap_exceptions: bool=False,
      strict: bool=True,
      commit: bool=True) -> Invocation[I_contra, O_co]:
    """Invoke the invokable, tracking invocation metadata.

    Args:
      arg: The input resource.
      replay_from: An optional invocation to replay form.
      exception_override: If replaying, an optional override for replayed
        exceptions.
      raise_on_errors: Which errors should raise beyond the invocation.
      wrap_exceptions: Whether to wrap non-resource exceptions in
        NativeException. If False, non-resource exceptions raised during
        invocation will be re-raised.
      strict: Whether replay should fail if the replayed invocation
        does not match the current invocation.
      commit: Whether to commit the new invocation object.
    Returns:
      The invocation generated.
    """
    arg = self._process_invoke_arg(arg)

    with self._invoke_exit_stack(replay_from, exception_override, strict):
      builder = Builder(self, arg)
      try:
        await builder.async_call()
      except raise_on_errors:
        raise
      except Exception:  # pylint: disable=broad-exception-caught
        should_raise_wrapped_error = (
          builder.exception_wrapped and not wrap_exceptions)
        if not builder.completed or should_raise_wrapped_error:
          raise
      invocation = builder.invocation
    if commit:
      references.commit(invocation)
    return invocation


@dataclasses.dataclass
class Invokable(InvokableBase[I_contra, O_co], resources.Resource):
  """Base class for dataclass-based invokable resources."""


@dataclasses.dataclass
class AsyncInvokable(AsyncInvokableBase[I_contra, O_co], resources.Resource):
  """Base class for dataclass-based async invokable resources."""


I = TypeVar('I', bound=_InvokableBase)


def typed_invokable(
    input_type: Optional[Type],
    output_type: Optional[Type]) -> Callable[
      [Type[I]], Type[I]]:
  """A decorator for creating typed invokables."""
  def _decorator(cls: Type[I]) -> Type[I]:
    """Decorates a class as a typed invokable."""
    if not issubclass(cls, _InvokableBase):
      raise TypeError('Invokable must be a subclass of InvokableBase.')
    # pylint: disable=protected-access
    cls._enact_input_type = input_type
    cls._enact_output_type = output_type
    resource_registry.register(cls)
    return cls
  return _decorator


@resource_registry.register
@dataclasses.dataclass
class RequestInput(Invokable):
  """An invokable that raises an InputRequest."""
  requested_type: Type
  context: Any = None

  def call(self, input_to_user: Any) -> Any:
    _request_input(input_to_user, self.requested_type, self.context)
    assert False


def request_input(
    requested_type: Type[AnyT],
    for_value: Any=None,
    context: Optional[interfaces.FieldValue]=None) -> AnyT:
  """Request an input from an external system / user.

  Args:
    requested_type: The type of input to request.
    for_value: The value to request input for. Defaults to None.
    context: An optional context to provide to the input request, e.g.,
      instructions to a user.
  Returns:
    The requested input. Note that this function will not return a value
    during normal execution, but will raise an InputRequest exception,
    which can be used to resume the execution with an injected value.
  Raises:
    InputRequest: Raised to halt execution in order to await input.
  """
  return RequestInput(requested_type, context)(for_value)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI bindings for enact invokables."""

import dataclasses
import inspect
from typing import List, Optional, Type

import enact


import fastapi

def get(
  app, route: str, invokable: enact.Invokable,
  store: enact.Store):
  """Exposes an envokable as a new get API endpoint.

  Args:
    app: The FastAPI app instance to modify.
    route: The route for the new API.
    invokable: The Invokable instance that the API is wrapping.
    store: The store instance that will hold the results of the invocation.
  """
  _add_fast_api(app, route, invokable, store, ['GET'])

def post(
  app, route: str, invokable: enact.Invokable,
  store: enact.Store):
  """Exposes an envokable as a new post API endpoint.

  Args:
    app: The FastAPI app instance to modify.
    route: The route for the new API.
    invokable: The Invokable instance that the API is wrapping.
    store: The store instance that will hold the results of the invocation.
  """
  _add_fast_api(app, route, invokable, store, ['POST'])

def _check_resource_type(t: Optional[Type[enact.ResourceBase]], in_out: str):
  """Check that input or output type is set and not overly general."""
  if t is None:
    raise ValueError(f'{in_out} type must be set on invokable, use the'
                     f'@enact.typed_invokable decorator')
  if t in (enact.Resource, enact.ResourceBase):
    raise ValueError(
      f'{in_out} type is overly general: {t}. FastAPI requires specific '
      f'types to be set for the request.')
  if issubclass(t, type(None)):
    return
  if not issubclass(t, enact.Resource):
    raise ValueError(
      f'{in_out} is not a subclass of enact.Resource: {t}. Only '
      f'dataclass-based resource inputs and outputs are supported by FastAPI.')
  dataclass_fields = set(f.name for f in dataclasses.fields(t))
  resource_fields = set(t.field_names())
  # Make sure that dataclass fields coincide with resource fields.
  if dataclass_fields != resource_fields:
    raise ValueError(
      f'{in_out} type {t} has dataclass fields {dataclass_fields} that do not '
      f'coincide with resource fields {resource_fields}. Using FastAPI '
      f'that all resource fields are declared as dataclass fields.')

def _check_invokable(invokable: enact.Invokable, methods: List[str]):
  """Check input and output type."""
  _check_resource_type(invokable.get_input_type(), 'Input')
  _check_resource_type(invokable.get_output_type(), 'Output')
  if 'GET' in methods and invokable.get_input_type() != type(None):
    raise ValueError(
      'GET is only supported for invokables with input type NoneResource.')

def _add_fast_api(
  app, route: str, invokable: enact.Invokable,
  store: enact.Store, methods: List[str]):
  """Internal implementation of wrapping an Invokable via FastAPI"""
  def get_output(invocation: enact.Invocation):
    if not invocation.successful():
      raise fastapi.HTTPException(
        status_code=500, detail=str(invocation.get_raised()))
    return invocation.get_output()

  def journaled_call() -> enact.Resource:
    with store:
      return get_output(invokable.invoke())

  def journaled_call_input(data: enact.Resource) -> enact.Resource:
    if data is None:
      raise fastapi.HTTPException(status_code=400, detail='No data provided')
    with store:
      return get_output(invokable.invoke(enact.commit(data)))

  _check_invokable(invokable, methods)

  param = inspect.Parameter(
    'data', inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None,
    annotation=invokable.get_input_type())

  journaled_call_input.__signature__ = inspect.Signature(  # type: ignore
    parameters=[param], return_annotation=invokable.get_output_type())

  wrapper_fn = (journaled_call if
                invokable.get_input_type() == type(None)
                else journaled_call_input)

  app.router.add_api_route(
    route, wrapper_fn, name=type(invokable).__name__, # type: ignore[arg-type]
    description=invokable.__doc__, response_model=invokable.get_output_type(),
    methods=methods)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default resource wrappers for non-field types."""

import dataclasses
import types as python_types
import sys
from typing import Type


from enact import resources
from enact import registration
from enact import types


@registration.register
@dataclasses.dataclass
class TupleWrapper(resources.TypeWrapper[tuple]):
  """Wrapper for tuples."""
  value: list

  @classmethod
  def is_immutable(cls) -> bool:
    return True

  @classmethod
  def wrapped_type(cls) -> Type[tuple]:
    return tuple

  @classmethod
  def wrap(cls, value: tuple) -> 'TupleWrapper':
    """Wrap a tuple value directly."""
    assert isinstance(value, tuple), (
      f'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return cls(list(value))

  def unwrap(self) -> tuple:
    """Unwrap the tuple."""
    return tuple(self.value)


@registration.register
@dataclasses.dataclass
class SetWrapper(resources.TypeWrapper[set]):
  """Wrapper for sets."""
  value: list

  @classmethod
  def wrapped_type(cls) -> Type[set]:
    return set

  @classmethod
  def wrap(cls, value: set) -> 'SetWrapper':
    """Wrap a set value directly."""
    assert isinstance(value, set), (
      f'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return cls(list(value))

  def unwrap(self) -> set:
    """Unwrap the set."""
    return set(self.value)


@registration.register
@dataclasses.dataclass
class TypeDescriptorWrapper(resources.TypeWrapper[types.TypeDescriptor]):
  """Wrapper for TypeDescriptors."""
  json: types.Json

  @classmethod
  def wrapped_type(cls) -> Type[types.TypeDescriptor]:
    return types.TypeDescriptor

  @classmethod
  def wrap(cls, value: types.TypeDescriptor) -> 'TypeDescriptorWrapper':
    """Wrap a type descriptor."""
    assert isinstance(value, types.TypeDescriptor), (
      f'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return cls(value.to_json())

  def unwrap(self) -> types.TypeDescriptor:
    """Unwrap the type descriptor."""
    return types.TypeDescriptor.from_json(self.json)


@registration.register
@dataclasses.dataclass
class ModuleWrapper(resources.TypeWrapper[python_types.ModuleType]):
  """Wrapper for python modules."""
  # TODO: Figure out a way to track type dependencies of a module wrapper.
  name: str

  @classmethod
  def wrapped_type(cls) -> Type[python_types.ModuleType]:
    """Returns the type of the wrapped resource."""
    # We could use any module here instead of sys, we just want to access
    # # <class 'module'>.
    return type(sys)

  @classmethod
  def wrap(cls, value: python_types.ModuleType) -> 'ModuleWrapper':
    """Wrap a python module."""
    assert isinstance(value, python_types.ModuleType), (
      f'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return cls(value.__name__)

  def unwrap(self) -> python_types.ModuleType:
    """Unwrap the module."""
    if self.name not in sys.modules:
      raise ValueError(
        f'Module {self.name} not found in sys.modules. Please make sure it '
        f'is imported before attempting to check out its reference.')
    return sys.modules[self.name]



# Copyright 2024 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Automatic field type inference."""

from typing import Any, Optional

from enact import interfaces, types, resource_registry


def from_annotation(t: Any) -> Optional[types.TypeDescriptor]:
  """Attempts to infer a type descriptor from a python type annotation."""
  if t == int:
    return types.Int()
  if t == str:
    return types.Str()
  if t == float:
    return types.Float()
  if t == bool:
    return types.Bool()
  if t == bytes:
    return types.Bytes()
  if t == list:
    return types.List()
  if t == dict:
    return types.Dict()
  if hasattr(t, '__origin__') and t.__origin__ in (types.List, list):
    return types.List()
  if hasattr(t, '__origin__') and t.__origin__ in (types.Dict, dict):
    return types.Dict()
  if isinstance(t, type):
    try:
      t = resource_registry.wrap_type(t)
    except resource_registry.MissingWrapperError:
      # We could raise here since this is not allowed, but we'll just ignore it
      # to not annoy users unless they need enact-based functionality, e.g.,
      # commits.
      return None
    if issubclass(t, interfaces.ResourceBase):
      return types.ResourceType(t.type_key())
  return None  # Could not infer type.



# Copyright 2024 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for utils."""

import unittest

from enact import types
from enact import resource_registry


class TypesTest(unittest.TestCase):
  """Tests for the types module."""

  def test_from_to_json(self):
    """Tests that the to_json and from_json methods work as expected."""
    for t in (types.Int(), types.Str(), types.Float(), types.Bool(),
              types.Bytes(), types.List(), types.Dict(), types.NoneType(),
              types.ResourceType(resource_registry.BoolWrapper.type_key())):
      got = types.TypeDescriptor.from_json(t.to_json())
      self.assertEqual(got, t)



# Copyright 2024 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the distribution registry."""

from importlib import metadata
import unittest
import re

import wrapt

import enact
from enact import distribution_registry
from enact import version


def _has_distribution(dist_name: str) -> bool:
  """Check if a distribution has been installed."""
  try:
    metadata.distribution(dist_name)
    return True
  except metadata.PackageNotFoundError:
    return False


def _is_editable_distribution(dist_name: str) -> bool:
  """Check if a distribution has been installed as editable."""
  expected_file_pattern = r'__editable__.*\.pth'
  files = metadata.distribution(dist_name).files
  return any(re.match(expected_file_pattern, str(f)) for f in files)


class DistributionRegistryTest(unittest.TestCase):
  """Tests for the distribution registry."""

  def test_explicit_registration(self):
    """Tests that explicit registration works."""
    d = distribution_registry.DistributionRegistry()
    d.register_distribution('foo', '1.0', '/path/to/foo')
    self.assertEqual(
      d.get_path_distribution_key('/path/to/foo/my.py'),
      enact.DistributionKey('foo', '1.0'))

  def test_register_editable_install(self):
    """Tests that editable installs can be registered."""
    # Check if enact is installed in editable mode.
    if (not _has_distribution(version.DIST_NAME)
        or not _is_editable_distribution(version.DIST_NAME)):
      self.skipTest('enact is not installed in editable mode')

    d = distribution_registry.DistributionRegistry()
    d.register_distribution(version.DIST_NAME)
    self.assertEqual(
      d.get_distribution_key(enact.ResourceBase),
      enact.DistributionKey(version.DIST_NAME, version.__version__))

  def test_register_non_editable_install(self):
    """Tests that editable installs can be registered."""
    # Check if wrapt is installed in non-editable mode.
    if (not _has_distribution('wrapt')
        or _is_editable_distribution('wrapt')):
      self.skipTest('wrapt is not installed in non-editable mode')

    d = distribution_registry.DistributionRegistry()
    d.register_distribution('wrapt')
    self.assertEqual(
      d.get_distribution_key(wrapt.decorator),
      enact.DistributionKey('wrapt', wrapt.__version__))



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tracks installation paths of python distributions."""

from importlib import metadata
import inspect
import os
from typing import Any, Dict, Iterable, List, Optional

from enact import types
from enact import version


_singleton: Optional['DistributionRegistry'] = None


def registry() -> 'DistributionRegistry':
  """Get the global package registry."""
  global _singleton
  if _singleton is None:
    _singleton = DistributionRegistry()
  assert _singleton is not None
  return _singleton


def _is_editable_pointer_file(dist_name: str, file_path: str) -> bool:
  """Check if this is the pointer file in the editable package."""
  base = os.path.basename(file_path)
  return (
    base.startswith('__editable__') and base.endswith('.pth')
    and dist_name in base)


class DistributionRegistry:
  """Keeps track of installed packages and their associated files."""

  def __init__(self):
    """Initialize a new package registry."""
    self._file_map: Dict[str, types.DistributionKey] = {}
    self._dir_map: Dict[str, types.DistributionKey] = {}
    self._registered: Dict[str, types.DistributionKey] = {}

  def registered(self) -> Iterable[types.DistributionKey]:
    """Yield the registered packages."""
    return self._registered.values()

  def is_registered(self, dist_name: str) -> bool:
    """Return whether the distribution is registered."""
    return dist_name in self._registered

  def register_distribution(
      self,
      dist_name: str,
      dist_version: Optional[str]=None,
      path: Optional[str]=None):
    """Register a distribution by the name you would use to install via pip.

    Args:
      dist_name: The name of the distribution.
      dist_version: The version of the distribution. If None, the function will
        attempt to infer the version via the installed package metadata.
      path: The install directory of distribution. If None, the function will
        attempt to directly register all files associated with the distribution
        with the provided package. For editable installs, this may be brittle
        and providing an explicit path is recommended.

    Raises:
      importlib.metadata.PackageNotFoundError: If the distribution is not
        installed and version and path were not explicitly provided.
    """
    files: List[str] = []
    if not dist_version or not path:
      dist = metadata.distribution(dist_name)
      if dist_version is None:
        dist_version = dist.metadata['version']
      if path is None and dist.files:
        files = [os.path.abspath(str(f.locate())) for f in dist.files]
    info = types.DistributionKey(
      name=dist_name, version=dist_version)
    if path:
      self._dir_map[path] = info
    for file_path in files:
      if _is_editable_pointer_file(dist_name, file_path):
        with open(file_path, 'r', encoding='utf-8') as file:
          path = os.path.abspath(file.read().strip())
          assert os.path.isdir(path), (
            f'Could not find path "{path}" to editable package '
            f'dist_name')
          self._dir_map[path] = info
      self._file_map[file_path] = info
    self._registered[dist_name] = info

  def get_path_distribution_key(self, path: str) -> (
      Optional[types.DistributionKey]):
    """Get the distribution key for a given path if it exists."""
    path = os.path.abspath(path)
    dist_info = self._file_map.get(path)
    if dist_info:
      return dist_info
    for dir_path, info in self._dir_map.items():
      if os.path.commonpath([dir_path, path]) == dir_path:
        return info
    return None

  def get_distribution_key(self, python_obj: Any) -> (
      Optional[types.DistributionKey]):
    """Get the distribution key for a given python type if it exists."""
    try:
      file = inspect.getfile(python_obj)
    except TypeError:
      # This happens for built-in types (or types that are wrongly
      # resolved as built-in types, such as types defined in ipynb notebook
      # cells).
      return None
    return self.get_path_distribution_key(file)

def register_distribution(
    dist_name: str,
    dist_version: Optional[str]=None,
    path: Optional[str]=None):
  """Register a distribution by the name you would use to install via pip.

  Args:
    dist_name: The name of the distribution.
    dist_version: The version of the distribution. If None, the function will
      attempt to infer the version via the installed package metadata.
    path: The install directory of distribution. If None, the function will
      attempt to directly register all files associated with the distribution
      with the provided package. For editable installs, this may be brittle
      and providing an explicit path is recommended.

  Raises:
    importlib.metadata.PackageNotFoundError: If the distribution is not
      installed and version and path were not explicitly provided.
  """
  return registry().register_distribution(dist_name, dist_version, path)

def get_path_distribution_key(path: str) -> (
    Optional[types.DistributionKey]):
  """Get the distribution key for a given path if it exists."""
  return registry().get_path_distribution_key(path)

def get_distribution_key(python_obj: Any) -> (
    Optional[types.DistributionKey]):
  """Get the distribution key for a given python type if it exists."""
  return registry().get_distribution_key(python_obj)

def ensure_enact_registered():
  """Ensure the enact distribution is registered."""
  # Register the current enact distribution.
  if registry().is_registered(version.DIST_NAME):
    return
  register_distribution(
    version.DIST_NAME,
    version.__version__,
    version.DIST_PYTHON_DIR)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the resource_registry module."""

import dataclasses
import os
import types
from typing import Any, List, Tuple, Type
import unittest
from unittest import mock

import enact
from enact import resource_registry
from enact import version
from enact import distribution_registry
from enact import types as enact_types


@enact.register
@dataclasses.dataclass
class SimpleResource(enact.Resource):
  """A simple resource for testing."""


@dataclasses.dataclass
class CustomType:
  """A custom non-resource type for testing."""
  tuple_value: Tuple


@enact.register
@dataclasses.dataclass
class CustomWrapper(enact.TypeWrapper[CustomType]):
  """A simple resource for testing."""
  list_value: list

  @classmethod
  def wrapped_type(cls) -> Type[CustomType]:
    return CustomType

  @classmethod
  def wrap(cls, wrapped: CustomType) -> 'CustomWrapper':
    return cls(list(wrapped.tuple_value))

  def unwrap(self) -> CustomType:
    return CustomType(tuple(self.list_value))


class RegistryTest(unittest.TestCase):
  """Tests the Registry module."""

  def test_registry(self):
    """Tests that the registry works."""
    registry = enact.Registry()
    registry.register(SimpleResource)
    self.assertEqual(registry.lookup(SimpleResource.type_id()), SimpleResource)

  def test_registry_error(self):
    """Tests that the registry raises an error for non-resources."""
    registry = enact.Registry()
    with self.assertRaises(enact.RegistryError):
      registry.register(int)  # type: ignore

  def test_lookup_error(self):
    """Tests that lookup raises the correct error."""
    registry = enact.Registry()
    with self.assertRaises(enact.ResourceNotFound):
      registry.lookup('SimpleResource')

  def test_singleton_and_decorator(self):
    """Tests the Registry singleton."""
    self.assertEqual(
      enact.Registry.get().lookup(SimpleResource.type_id()),
      SimpleResource)

  def test_wrap_field_values(self):
    """Tests wrapping and unwrapping field values."""
    field_values = [
      (None, type(None)), (1, int), (1.0, float), ('a', str),
      (bytes([1, 2, 3]), bytes), ([1, 2, 3], list),
      ({'1': 1, '2': 2}, dict)]
    for value, value_type in field_values:
      with self.subTest(str(value_type)):
        wrapped = enact.wrap(value)
        assert isinstance(wrapped, enact.TypeWrapperBase)
        self.assertEqual(enact.unwrap(wrapped), value)

  def test_wrap_noop_on_resources(self):
    """Tests that wrapping does nothing on resources."""
    resource = SimpleResource()
    self.assertEqual(enact.wrap(resource), resource)

  def test_wrap_custom_type(self):
    """Tests wrapping custom types."""
    custom = CustomType((1, 2, 3))
    wrapped = enact.wrap(custom)
    assert isinstance(wrapped, CustomWrapper)
    self.assertEqual(enact.unwrap(wrapped), custom)

  def test_wrap_nests(self):
    """Tests that wrapping nests with custom types works."""
    py_dict = {'a': [CustomType((1, 2)), {'test': CustomType((2, 3))}]}
    wrapped = enact.wrap(py_dict)
    assert isinstance(wrapped, enact.TypeWrapperBase)
    self.assertEqual(enact.unwrap(wrapped), py_dict)

  def test_wrap_types(self):
    """Tests that wrapping types works."""
    type_nest = [int, [float, bytes]]
    as_fields = resource_registry.to_field_value(type_nest)
    self.assertEqual(
      as_fields, [resource_registry.IntWrapper,
                  [resource_registry.FloatWrapper,
                   resource_registry.BytesWrapper]])
    self.assertEqual(
      resource_registry.from_field_value(as_fields), type_nest)

  def test_wrap_function(self):
    """Tests that wrapping functions works."""
    def foo(x: int) -> int:
      return x

    def foo_wrapper(x: int) -> int:
      return x

    @dataclasses.dataclass
    class MethodWrapper(resource_registry.MethodWrapper):
      @classmethod
      def wrapper_function(cls):
        return foo_wrapper

      @classmethod
      def wrapped_function(cls):
        return foo

    @dataclasses.dataclass
    class FunctionWrapper(resource_registry.FunctionWrapper):
      """The FunctionWrapper that goes with MethodWrapper."""
      @classmethod
      def wrapper_function(cls):
        return foo_wrapper

      @classmethod
      def wrapped_function(cls):
        return foo

      @classmethod
      def method_wrapper(cls) -> Type[resource_registry.MethodWrapper]:
        return MethodWrapper

    registry = enact.Registry()
    registry.register(FunctionWrapper)
    self.assertEqual(registry.wrap(foo_wrapper), FunctionWrapper())
    self.assertEqual(
      registry.unwrap(registry.wrap(foo_wrapper)), foo_wrapper)

  def test_wrap_method(self):
    """Tests that wrapping methods works."""

    @dataclasses.dataclass
    class MyClass(enact.Resource):
      y: int

      def foo(self, x: int) -> int:
        return x

      def foo_wrapper(self, x: int) -> int:
        return x

    @dataclasses.dataclass
    class MethodWrapper(resource_registry.MethodWrapper):
      instance: Any

      @classmethod
      def wrapper_function(cls):
        return MyClass.foo_wrapper

      @classmethod
      def wrapped_function(cls):
        return MyClass.foo

    @dataclasses.dataclass
    class FunctionWrapper(resource_registry.FunctionWrapper):
      """The FunctionWrapper that goes with MethodWrapper."""
      @classmethod
      def wrapper_function(cls):
        return MyClass.foo_wrapper

      @classmethod
      def wrapped_function(cls):
        return MyClass.foo

      @classmethod
      def method_wrapper(cls) -> Type[resource_registry.MethodWrapper]:
        return MethodWrapper

    registry = enact.Registry()
    registry.register(MyClass)
    registry.register(FunctionWrapper)

    self.assertEqual(registry.wrap(MyClass.foo_wrapper), FunctionWrapper())
    self.assertEqual(
      registry.unwrap(registry.wrap(MyClass.foo_wrapper)), MyClass.foo_wrapper)

    instance = MyClass(69)
    wrapped_method = registry.wrap(instance.foo_wrapper)
    self.assertEqual(wrapped_method, MethodWrapper(instance))
    unwrapped_method = registry.unwrap(wrapped_method)
    self.assertIsInstance(unwrapped_method, types.MethodType)
    self.assertEqual(unwrapped_method.__self__, instance)
    self.assertEqual(unwrapped_method.__func__, MyClass.foo_wrapper)

    # Check that original instance and self are aliased.
    instance.y = 420
    self.assertEqual(unwrapped_method.__self__.y, 420)

  def test_wrap_type_value(self):
    """Tests that wrapping methods works."""
    @enact.register
    class MyResource(enact.Resource):
      pass

    # Test for resource subclasses.
    self.assertIsInstance(
      enact.wrap(MyResource),
      resource_registry.ResourceTypeWrapper)

    # Test for types that have associated wrappers.
    self.assertIsInstance(
      enact.wrap(int),
      resource_registry.ResourceTypeWrapper)

    with enact.Store():
      resource_type_ref = enact.commit(MyResource)
      int_type_ref = enact.commit(int)

      # Clear cache
      resource_type_ref = enact.Ref.from_id(resource_type_ref.id)
      int_type_ref = enact.Ref.from_id(int_type_ref.id)

      # Checkout
      self.assertEqual(resource_type_ref.checkout(), MyResource)
      self.assertEqual(int_type_ref.checkout(), int)

  def test_wrap_unregistered_type_fails(self):
    """Tests that wrapping an unregistered type fails."""
    class MyUnregisteredClass:
      pass
    with enact.Store():
      with self.assertRaises(resource_registry.MissingWrapperError):
        enact.wrap(MyUnregisteredClass)

  def test_deepcopy_primitives(self):
    """Tests that deep copying primitives works."""
    primitives = [1, 1.0, 'a', bytes([1, 2, 3]), True, False, None]
    for p in primitives:
      with self.subTest(str(p)):
        self.assertEqual(resource_registry.deepcopy(p), p)

  def test_deepcopy_nests(self):
    """Tests that deep copying primitives works."""
    nest: List[Any] = [[1, 2], {'a': [1, 2], 'b': True}]
    copy = resource_registry.deepcopy(nest)
    self.assertEqual(copy, nest)
    self.assertIsNot(copy, nest)
    self.assertIsNot(copy[0], nest[0])
    self.assertIsNot(copy[1], nest[1])
    self.assertIsNot(copy[1]['a'], nest[1]['a'])

class DistributionInfoTests(unittest.TestCase):
  """Tests the distribution info related functionality."""

  def test_enact_types_have_distribution_key(self):
    """Tests that type distribution key is present for enact types."""
    self.assertEqual(
      resource_registry.IntWrapper.type_distribution_key(),
      enact.DistributionKey(version.DIST_NAME, version.__version__))

  def test_auto_assign_distribution_key(self):
    """Tests that type distribution key is auto-assigned on register"""
    dist_registry = distribution_registry.DistributionRegistry()

    # Patch out the singleton distribution registry.
    with mock.patch.object(
        distribution_registry, 'registry',  lambda: dist_registry):
      dist_registry.register_distribution(
        'enact-tests', '0.0.1', os.path.dirname(__file__))
      class MyResource(enact.Resource):
        pass
      self.assertIsNone(MyResource.type_distribution_key())
      enact.register(MyResource)
      self.assertEqual(
        MyResource.type_distribution_key(),
        enact.DistributionKey('enact-tests', '0.0.1'))

class TestToPythonType(unittest.TestCase):
  """Tests casting type descriptors to python types."""

  def test_to_from_python_type(self):
    """Tests that type descriptors can be cast to python types."""
    @enact.register
    class MyResource(enact.Resource):
      pass

    @dataclasses.dataclass  # dataclass for value comparison.
    class MyWrappedType:
      pass

    @enact.register
    class MyWrapperType(enact.TypeWrapper[MyWrappedType]):
      """Wrapper for testing."""
      @classmethod
      def wrap(cls, unused_wrapped: MyWrappedType) -> 'MyWrapperType':
        return cls()
      @classmethod
      def wrapped_type(cls) -> Type[MyWrappedType]:
        return MyWrappedType
      def unwrap(self) -> MyWrappedType:
        return MyWrappedType()

    tests = {
      int: enact_types.Int(),
      str: enact_types.Str(),
      float: enact_types.Float(),
      bool: enact_types.Bool(),
      bytes: enact_types.Bytes(),
      list: enact_types.List(),
      dict: enact_types.Dict(),
      MyResource: enact_types.ResourceType(MyResource.type_key()),
      MyWrappedType: enact_types.ResourceType(MyWrapperType.type_key()),
    }

    for python_type, descriptor in tests.items():
      with self.subTest(python_type=python_type):
        self.assertEqual(
          resource_registry.to_python_type(descriptor), python_type)
        self.assertEqual(
          resource_registry.from_python_type(python_type), descriptor)


if __name__ == 'main':
  unittest.main()



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for resources."""


import dataclasses
from typing import Any, Dict, List, Optional, Type, Union
import unittest

import enact
from enact import resource_registry
from enact import types


@enact.register
@dataclasses.dataclass
class SimpleResource(enact.Resource):
  a: Any
  b: Any
  c: Any


@enact.register
@dataclasses.dataclass
class OtherSimpleResource(enact.Resource):
  a: Any
  b: Any
  c: Any


class ResourceTest(unittest.TestCase):
  """Tests for the resource protocol."""

  def test_field_names(self):
    """Tests that the fields are correct."""
    self.assertEqual(
      list(SimpleResource.field_names()),
      ['a', 'b', 'c'])

  def test_field_values(self):
    """Tests that the field values are correct."""
    r = SimpleResource(1, 2, 3)
    self.assertEqual(
      list(r.field_values()),
      [1, 2, 3])

  def test_field_items(self):
    """Tests that field_items works."""
    r = SimpleResource(1, 2, 3)
    self.assertEqual(
      list(r.field_items()),
      [('a', 1), ('b', 2), ('c', 3)])

  def test_type_id_same_fields(self):
    """Tests that the type id is correct."""
    @dataclasses.dataclass
    class R(enact.Resource):
      a: Any
    type_id = R.type_id()
    # pylint: disable=function-redefined
    @dataclasses.dataclass
    class R(enact.Resource):  # type: ignore
      a: Any
    self.assertEqual(type_id, R.type_id())

  def test_field_wrapping(self):
    """Tests field wrapping."""
    class Custom:
      """A custom class."""
      def __init__(self, x: int):
        self.x = x

    @enact.register
    @dataclasses.dataclass
    class CustomWrapper(enact.TypeWrapper[Custom]):
      """A wrapper for Custom."""
      x: int
      @classmethod
      def wrap(cls, wrapped: Custom) -> 'CustomWrapper':
        return cls(wrapped.x)

      def unwrap(self) -> Custom:
        return Custom(self.x)

      @classmethod
      def wrapped_type(cls) -> Type[Custom]:
        return Custom

    @enact.register
    @dataclasses.dataclass
    class CustomFieldsResource(enact.Resource):
      """Tests a resource with custom field types."""
      w: Custom
      i: int

    r = CustomFieldsResource(Custom(1), 2)
    as_dict = r.to_resource_dict()
    w_dict = as_dict['w']
    assert isinstance(w_dict, enact.ResourceDict)
    self.assertEqual(w_dict.type_info, CustomWrapper.type_key())
    rebuilt = resource_registry.from_resource_dict(as_dict)
    self.assertEqual(r.w.x, rebuilt.w.x)
    self.assertEqual(r.i, rebuilt.i)

class RefTest(unittest.TestCase):
  """Tests for references."""

  def test_digest_identical(self):
    """Test that digest is identical for identical resources."""
    a = SimpleResource(1, 2, 3)
    b = SimpleResource(1, 2, 3)
    _, r1 = enact.Ref.pack(a)
    _, r2 = enact.Ref.pack(b)
    self.assertEqual(r1, r2)

  def test_typename_changes_hash(self):
    """Tests that changing the typename changes the hash."""
    a = SimpleResource(1, 2, 3)
    b = OtherSimpleResource(1, 2, 3)
    _, r1 = enact.Ref.pack(a)
    _, r2 = enact.Ref.pack(b)
    self.assertNotEqual(r1.ref(), r2.ref())

  def test_value_changes_hash(self):
    """Tests that changing values changes the hash."""
    a = SimpleResource(1, 2, 3)
    b = SimpleResource(1, 2, 4)
    _, r1 = enact.Ref.pack(a)
    _, r2 = enact.Ref.pack(b)
    self.assertNotEqual(r1.ref(), r2.ref())

  def test_hash_complex_nested(self):
    """Tests that nested resources are hashed correctly."""
    a = SimpleResource(
      [1, None, 2.0, True, False],
      [{'a': {'b': {'c': 1}}}],
      enact.Ref('1234'))
    b = SimpleResource(
      [1, None, 2.0, True, False],
      [{'a': {'b': {'c': 1}}}],
      enact.Ref('1234'))
    self.assertEqual(
      enact.Ref.pack(a),
      enact.Ref.pack(b))

  def test_nested_resource(self):
    """Tests nesting resources."""
    a = SimpleResource(SimpleResource(1, 2, 3), 2, 3)
    enact.Ref.pack(a)

  def test_unknown_field_error(self):
    """Tests that including an unknown type will raise an error."""
    class X:
      pass
    a = SimpleResource(X(), 2, 3)
    with self.assertRaises(enact.FieldTypeError):
      enact.Ref.pack(a)

  def test_deep_copy_resource(self):
    """Tests that the resource can be deep-copied."""
    a = SimpleResource(SimpleResource(1, 2, 3), [4, None],
                       {'a': 0.0, 'b': [True, False]})
    b = enact.deepcopy(a)
    self.assertEqual(a, b)
    self.assertNotEqual(id(a), id(b))
    self.assertNotEqual(id(a.a), id(b.a))
    self.assertNotEqual(id(a.b), id(b.b))
    self.assertNotEqual(id(a.c), id(b.c))
    self.assertEqual(enact.Ref.pack(a), enact.Ref.pack(b))

  def test_set_from(self):
    """Tests that set-from works as expected."""
    x = SimpleResource(SimpleResource(1, 2, 3), [4, None],
                       {'a': 0.0, 'b': [True, False]})
    y = SimpleResource(None, None, None)
    y.set_from(x)
    self.assertEqual(y, x)
    x.a.a = 5
    self.assertNotEqual(y, x)


class MyClass:
  """Test non-resource class"""
  pass


@enact.register
class MyClassWrapper(enact.TypeWrapper[MyClass]):
  """Wrapper for MyClass"""
  @classmethod
  def wrap(cls, unused_wrapped: MyClass) -> 'MyClassWrapper':
    return cls()

  @classmethod
  def wrapped_type(cls) -> Type[MyClass]:
    return MyClass

  def unwrap(self) -> MyClass:
    return MyClass()


@dataclasses.dataclass
class TypeExampleResource(enact.Resource):
  """Test resource."""
  bool_value: bool
  bytes_value: 'bytes'
  str_value: str
  float_value: float
  int_value: int
  list_value: list
  list_value_generic: List[int]
  dict_value: dict
  dict_value_generic: Dict[str, int]
  a_value: 'TypeExampleResource'
  a_list: List['TypeExampleResource']
  a_dict: 'Dict[str, TypeExampleResource]'
  union_value: Union[int, float]
  optional_value: Optional[int]
  wrapped_value: 'MyClass'


class TypeDescriptorTests(unittest.TestCase):
  """Tests automatic resolution of type descriptors."""

  def test_field_descriptors(self):
    """Tests that field descriptors are inferred correctly."""
    descriptors = list(TypeExampleResource.field_descriptors())
    self.assertEqual(descriptors, [
      types.Bool(),
      types.Bytes(),
      types.Str(),
      types.Float(),
      types.Int(),
      types.List(),
      types.List(),
      types.Dict(),
      types.Dict(),
      types.ResourceType(
        TypeExampleResource.type_key()),
      types.List(),
      types.Dict(),
      None,  # Union not supported
      None,  # Optional not supported
      types.ResourceType(MyClassWrapper.type_key())
    ])



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass-based resources."""

import abc
import dataclasses
from typing import Iterable, Mapping, Optional, Type, TypeVar
import typing

from enact import interfaces
from enact import resource_registry
from enact import types
from enact import type_inference


FieldValue = interfaces.FieldValue

C = TypeVar('C', bound='_Resource')


class _Resource(interfaces.ResourceBase):
  """Base class for Resource and FrozenResource."""

  @classmethod
  def field_names(cls) -> Iterable[str]:
    """Returns the names of the fields of the resource."""
    return (f.name for f in dataclasses.fields(cls))  # type: ignore

  def field_values(self) -> Iterable[FieldValue]:
    """Return a list of field values, aligned with field_names."""
    return (resource_registry.to_field_value(
      getattr(self, f)) for f in self.field_names())

  @classmethod
  def field_descriptors(cls) -> Iterable[Optional[types.TypeDescriptor]]:
    """Infer field descriptors from type annotations where possible."""
    # We use get_type_hints since it resolves types specified as strings,
    # whereas dataclasses.fields does not.
    try:
      type_dict = typing.get_type_hints(cls)
    except NameError:
      # This can happen for some type definitions.
      return (None for _ in cls.field_names())
    return (
      type_inference.from_annotation(type_dict.get(name))
      for name in cls.field_names())

  @classmethod
  def from_fields(cls: Type[C],
                  field_values: Mapping[str, FieldValue]) -> C:
    """Constructs the resource from a value dictionary."""
    field_values = {k: resource_registry.from_field_value(v)
                    for k, v in field_values.items()}
    return cls(**field_values)



ResourceT = TypeVar('ResourceT', bound='Resource')

@dataclasses.dataclass
class Resource(_Resource):
  """Base class for dataclass-based resources.

  Subclasses must be registered with the @enact.register decorator in order
  to allow deserialization from references.
  """
  def set_from(self, other: interfaces.ResourceBase):
    """Sets the fields of this resource from another resource.

    Implementation of set_from is required to support replays of invokable
    resources that change their internal state during execution.
    """
    if not type(self) == type(other):  # pylint: disable=unidiomatic-typecheck
      raise TypeError(f'Cannot set_from {type(other)} into {type(self)}.')
    copy = resource_registry.deepcopy(other)
    for field in dataclasses.fields(self):
      self_field = getattr(self, field.name)
      target = getattr(other, field.name)
      if self_field is target:
        continue
      if type(self_field) == type(target):  # pylint: disable=unidiomatic-typecheck
        if isinstance(self_field, interfaces.ResourceBase):
          # In cases where we have compatible field values, we can set them
          # directly without creating a new instance. Note that this may lead to
          # unexpected behavior in cases where a new object is explicitly
          # allocated (and later checked, e.g., via ID). See
          # https://github.com/agentic-ai/enact/issues/54
          self_field.set_from(target)
          continue
        else:
          wrapper_type = resource_registry.Registry.get().get_type_wrapper(
            type(self_field))
          if wrapper_type and not wrapper_type.is_immutable():
            wrapper_type.set_wrapped_value(self_field, target)
            continue
      setattr(self, field.name, getattr(copy, field.name))


@dataclasses.dataclass(frozen=True)
class ImmutableResource(_Resource):
  """Base class for immutable dataclass-based resources.

  Subclasses must be registered with the @enact.register decorator in order
  to allow deserialization from references.
  """
  def set_from(self, other: interfaces.ResourceBase):
    """Sets the fields of this resource from another resource.

    Implementation of set_from is required to support replays of invokable
    resources that change their internal state during execution.
    """
    if self != other:
      raise TypeError(f'Cannot call set_from on immutable resource {self}.')


WrappedT = TypeVar('WrappedT')
WrapperT = TypeVar('WrapperT', bound='TypeWrapper')


@dataclasses.dataclass
class TypeWrapper(interfaces.TypeWrapperBase[WrappedT], Resource):
  """Base class for dataclass-based type wrappers."""

  @classmethod
  @abc.abstractmethod
  def wrapped_type(cls) -> Type[WrappedT]:
    """Returns the type wrapped by this TypeWrapper."""
    raise NotImplementedError()

  @classmethod
  @abc.abstractmethod
  def wrap(cls: Type[WrapperT], value: WrappedT) -> WrapperT:
    """Wrap a value."""
    raise NotImplementedError()

  @abc.abstractmethod
  def unwrap(self) -> WrappedT:
    """Wrap a value."""
    raise NotImplementedError()



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization functionality."""

import abc
import base64
import json
from typing import Dict, Mapping, Optional, Sequence, Union, cast

from enact import interfaces
from enact import resource_registry


JsonLeaf = Union[int, float, str, bool, None]
Json = Union[JsonLeaf, Sequence['Json'], Dict[str, 'Json']]

JSON_LEAF_TYPES = (int, float, str, bool, type(None))


class SerializationError(Exception):
  """Raised when serialization fails."""


class DeserializationError(Exception):
  """Raised when deserialization fails."""


class Serializer(abc.ABC):
  """Serializes and deserializes resources."""

  @abc.abstractmethod
  def serialize(self, resource_dict: interfaces.ResourceDict) -> bytes:
    """Serializes a resource."""

  @abc.abstractmethod
  def deserialize(self, data: bytes) -> interfaces.ResourceDict:
    """Deserializes a packed resource."""


class JsonSerializer(Serializer):
  """A JSON serializer."""

  _ESCAPE = '#'

  def __init__(self,
               registry: Optional[resource_registry.Registry]=None,
               encoding='utf-8'):
    """Initializes a JSON serializer."""
    self._registry = registry or resource_registry.Registry.get()
    self._encoding = encoding

  def _escape(self, s: str):
    """Escapes a string."""
    return f'{self._ESCAPE}{s}'

  def to_json(self, value: interfaces.ResourceDictValue) -> Json:
    """Converts a value to a JSON compatible value recursively."""
    if isinstance(value, interfaces.ResourceDict):
      result: Dict[str, Json] = {
        self._escape('res'): (value.type_info.type_id())}
      for k, v in value.items():
        if not isinstance(k, str):
          raise interfaces.FieldTypeError(
            f'Resource has non-string keys: {value}')
        if k.startswith(self._ESCAPE):
          raise SerializationError(
            f'Cannot serialize resource with fields starting with '
            f'{self._ESCAPE}: {value}')
        result[k] = self.to_json(v)
      return result
    if isinstance(value, JSON_LEAF_TYPES):
      return value
    if isinstance(value, bytes):
      return {
        self._escape('b85'): base64.b85encode(value).decode('ascii')}
    if isinstance(value, type):
      if not issubclass(value, interfaces.ResourceBase):
        raise SerializationError(
          f'Cannot serialize type: {value} which is not a resource')
      return {
        self._escape('type'): value.type_id()}
    if isinstance(value, Mapping):
      result = {}
      for k, v in value.items():
        if not isinstance(k, str):
          raise interfaces.FieldTypeError(
            f'Mapping has non-string keys: {value}')
        if k.startswith(self._ESCAPE):
          raise SerializationError(
            f'Cannot serialize mapping with keys starting with '
            f'{self._ESCAPE}: {value}')
        result[k] = self.to_json(v)
      return result
    if isinstance(value, Sequence):
      return [self.to_json(v) for v in value]
    raise SerializationError(
      f'Cannot serialize value: {value} of type {type(value)}')

  def _resource_dict_from_json(
      self,
      value: Mapping[str, Json]) -> interfaces.ResourceDict:
    """Parse a resource from a json representation."""
    type_id_key = self._escape('res')
    fields = dict(value)
    type_id = fields.pop(type_id_key, None)
    if not type_id:
      raise DeserializationError(
        f'Could not find {type_id_key} in JSON: {value}')
    if not isinstance(type_id, str):
      raise DeserializationError(
        f'Expected {type_id_key} to be a string, '
        f'got {type(type_id)}: {value}')
    resource_class = self._registry.lookup(type_id)
    return interfaces.ResourceDict(resource_class, self.from_json(fields))

  def from_json(self, value: Json) -> interfaces.ResourceDictValue:
    """Turn a json encodable dictionary into a field value."""
    # Check encoded resource:
    type_id_key = self._escape('res')
    b85_key = self._escape('b85')
    resource_type_key = self._escape('type')
    if isinstance(value, JSON_LEAF_TYPES):
      return value
    if isinstance(value, Mapping):
      if type_id_key in value:
        ref = self._resource_dict_from_json(value)
        return ref
      if b85_key in value:
        b85str = cast(str, value[b85_key])
        return base64.b85decode(b85str)
      if resource_type_key in value:
        type_id = cast(str, value[resource_type_key])
        return self._registry.lookup(type_id)
      return {k: self.from_json(v) for k, v in value.items()}
    elif isinstance(value, Sequence):
      return [self.from_json(v) for v in value]
    else:
      raise DeserializationError(
        f'Cannot deserialize value: {value} of type {type(value)}')

  def serialize(self, resource_dict: interfaces.ResourceDict) -> bytes:
    """Serializes a resource dictionary."""
    json_dict = self.to_json(resource_dict)
    return json.dumps(json_dict, sort_keys=True).encode(self._encoding)

  def deserialize(self, data: bytes) -> interfaces.ResourceDict:
    """Deserializes a resource."""
    json_dict = json.loads(data.decode(self._encoding))
    assert isinstance(json_dict, dict)
    return self._resource_dict_from_json(json_dict)



# Copyright 2024 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A serializable type description language."""

import abc
import dataclasses
import typing
import json


JsonLeaf = typing.Union[int, float, str, bool, None]
Json = typing.Union[JsonLeaf, typing.List['Json'], typing.Dict[str, 'Json']]

PRIMITIVES = (int, float, str, bytes, bool, type(None))

Primitives = typing.Union[
  int, float, str, bytes, bool, None]


class TypeKey(typing.NamedTuple):
  """Information about a resource type."""
  name: str
  distribution_key: typing.Optional['DistributionKey']

  def type_id(self) -> str:
    """Returns a unique string identifier for the type."""
    return json.dumps(self.as_dict(), sort_keys=True)

  def as_dict(self) -> typing.Dict[str, Json]:
    """Returns a dictionary representation of the distribution key."""
    dist_info = (
      self.distribution_key.as_dict() if self.distribution_key else None)
    return {
      'name': self.name,
      'distribution_key': dist_info}

  @staticmethod
  def from_dict(d: typing.Dict[str, Json]) -> 'TypeKey':
    """Instantiate TypeKey from a dictionary."""
    r: typing.Dict = dict(d)
    r['distribution_key'] = DistributionKey.from_dict(r['distribution_key'])
    return TypeKey(**r)


class DistributionKey(typing.NamedTuple):
  """Information about a package where a resource is defined.

  This information (together with qualified class names) is used to identify
  compatible resources across different versions of a package.
  """
  name: str
  version: str

  def as_dict(self) -> typing.Dict[str, Json]:
    """Returns a dictionary representation of the distribution key."""
    return {'name': self.name, 'version': self.version}

  @staticmethod
  def from_dict(d: typing.Dict[str, Json]) -> 'DistributionKey':
    """Instantiate DistributionKey from a dictionary."""
    return DistributionKey(**typing.cast(typing.Dict[str, str], d))


@dataclasses.dataclass  # Type descriptors are dataclasses for value-semantics.
class TypeDescriptor(abc.ABC):
  """Interface for type descriptors."""
  NAME: typing.ClassVar[str] = ''

  def to_json(self) -> Json:
    """JSON representation."""
    return self.NAME

  @staticmethod
  def from_json(json_value: Json) -> 'TypeDescriptor':
    """Constructs a descriptor from JSON."""
    cls: typing.Type[TypeDescriptor]
    for cls in BASIC_TYPE_DESCRIPTOR_CLASSES:
      if json_value == cls.NAME:
        return cls()
    if not isinstance(json_value, dict) or len(json_value) != 1:
      raise ValueError(
        f'Expected a dictionary of length 1 for complex type descriptor: '
        f'{json_value}')
    if ResourceType.NAME in json_value:
      value = json_value[ResourceType.NAME]
      if not isinstance(value, dict):
        raise ValueError(
          f'Expected a dictionary for resource type descriptor: '
          f'{json_value}')
      return ResourceType(TypeKey.from_dict(value))
    raise ValueError(f'Unknown type descriptor: {json_value}')


@dataclasses.dataclass
class Int(TypeDescriptor):
  """Describes an int value."""
  NAME: typing.ClassVar[str] = 'int'


@dataclasses.dataclass
class Float(TypeDescriptor):
  """Describes a float value."""
  NAME: typing.ClassVar[str] = 'float'


@dataclasses.dataclass
class Str(TypeDescriptor):
  """Describes a string value."""
  NAME: typing.ClassVar[str] = 'str'


@dataclasses.dataclass
class Bool(TypeDescriptor):
  """Describes a boolean value."""
  NAME: typing.ClassVar[str] = 'bool'


@dataclasses.dataclass
class Bytes(TypeDescriptor):
  """Describes a bytes value."""
  NAME: typing.ClassVar[str] = 'bytes'


@dataclasses.dataclass
class List(TypeDescriptor):
  """Describes a list value."""
  NAME: typing.ClassVar[str] = 'list'


@dataclasses.dataclass
class Dict(TypeDescriptor):
  """Describes a dictionary value."""
  NAME: typing.ClassVar[str] = 'dict'


@dataclasses.dataclass
class NoneType(TypeDescriptor):
  """Describes a None type."""
  NAME: typing.ClassVar[str] = 'none'


@dataclasses.dataclass
class ResourceType(TypeDescriptor):
  """Describes a resource type."""
  NAME: typing.ClassVar[str] = 'resource'
  type_key: TypeKey

  def to_json(self):
    return {self.NAME: self.type_key.as_dict()}


BASIC_TYPE_DESCRIPTOR_CLASSES = (
  Int, Float, Str, Bool, Bytes, List, Dict, NoneType)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes hash digests of resources."""

import hashlib
from typing import Any, Iterable, List, Mapping, Sequence, Tuple, Type, Union

from enact import interfaces


Value = Union[
  interfaces.FieldValue,
  interfaces.ResourceDictValue]


def type_digest(cls: Type[interfaces.ResourceBase]) -> str:
  """Return a digest of the type of resource."""
  hash_obj = hashlib.sha256()
  hash_obj.update(cls.__module__.encode('utf-8'))
  hash_obj.update(b'.')
  hash_obj.update(cls.__qualname__.encode('utf-8'))
  for field in sorted(cls.field_names()):
    hash_obj.update(repr(field).encode('utf-8'))
  return hash_obj.hexdigest()


def _digest(
    value: Value,
    hash_obj: Any,
    stack: List[int]):
  """Recursively compute digest over a field value.

  Args:
    value: The value to digest.
    hash: A hash object with an update method, e.g., hashlib.sha256().
  """
  if id(value) in stack:
    raise interfaces.FieldTypeError(
      'Cyclic references are not allowed in field values.')
  stack.append(id(value))
  if isinstance(value, (interfaces.ResourceBase,
                        interfaces.ResourceDict)):
    items: Iterable[Tuple[str, Value]]
    if isinstance(value, interfaces.ResourceBase):
      type_id = type(value).type_id()
      # Use alphabetical ordering.
      items = sorted(value.field_items(), key=lambda x: x[0])
    else:
      assert isinstance(value, interfaces.ResourceDict)
      type_id = value.type_info.type_id()
      items = sorted(value.items(), key=lambda x: x[0])
    hash_obj.update(b'res[')
    hash_obj.update(type_id.encode('utf-8'))
    for k, v in items:
      hash_obj.update(repr(k).encode('utf-8'))
      _digest(v, hash_obj, stack)
    hash_obj.update(b']')
  elif isinstance(value, int):
    hash_obj.update(b'i')
    hash_obj.update(repr(int(value)).encode('utf-8'))
  elif isinstance(value, float):
    hash_obj.update(b'f')
    hash_obj.update(repr(value).encode('utf-8'))
  elif isinstance(value, str):
    hash_obj.update(b's')
    hash_obj.update(repr(value).encode('utf-8'))
  elif isinstance(value, bytes):
    hash_obj.update(b'b')
    hash_obj.update(value)
  elif value is True:
    hash_obj.update(b'1')
  elif value is False:
    hash_obj.update(b'0')
  elif value is None:
    hash_obj.update(b'n')
  elif isinstance(value, Sequence):
    hash_obj.update(b'seq[')
    for item in value:
      _digest(item, hash_obj, stack)
    hash_obj.update(b']')
  elif isinstance(value, Mapping):
    hash_obj.update(b'map[')
    for k, v in sorted(value.items(), key=lambda x: x[0]):  # type: ignore
      if not isinstance(k, str):
        raise interfaces.FieldTypeError('Map keys must be strings')
      hash_obj.update(repr(k).encode('utf-8'))
      _digest(v, hash_obj, stack)
    hash_obj.update(b']')
  elif issubclass(value, interfaces.ResourceBase):
    # Type of resource.
    hash_obj.update(b'type[')
    hash_obj.update(value.type_id().encode('utf-8'))
    hash_obj.update(b']')
  else:
    raise interfaces.FieldTypeError(
      f'Got unexpected field type: {type(value)}. '
      f'Allowed fields types are: int, float, str, bytes, bool, None and '
      f'Ref, and nested maps from strings or sequences of these types.')
  stack.pop()


def digest(resource: Union[interfaces.ResourceDict,
                           interfaces.ResourceBase]) -> str:
  """Compute a digest of a resource or a dict representation."""
  hash_obj = hashlib.sha256()
  _digest(resource, hash_obj, [])
  return hash_obj.hexdigest()



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for references and stores."""

import dataclasses
import tempfile
from typing import List
import unittest

import enact
from enact import interfaces
from enact import references
from enact import serialization
from enact import contexts
from enact import resource_registry
from enact import acyclic
from enact import type_wrappers
from enact import types
from enact import version

# pylint: disable=invalid-name,missing-class-docstring

@enact.register
@dataclasses.dataclass
class SimpleResource(enact.Resource):
  x: int
  y: float


@enact.register
@dataclasses.dataclass
class JsonPackedResource(enact.Resource):
  contents: bytes


@enact.register
class JsonPackedRef(enact.Ref):
  """A reference to a JSON packed resource."""

  @classmethod
  def verify(cls, packed_resource: references.PackedResource):
    """Verifies that the packed resource is valid."""

  @classmethod
  def unpack(cls, packed_resource: references.PackedResource) -> (
      interfaces.ResourceBase):
    """Unpacks the referenced resource."""
    data = packed_resource.data
    if data.type_info != JsonPackedResource.type_key():
      raise enact.RefError('Resource is not a JsonPackedResource.')
    json_packed: JsonPackedResource = resource_registry.from_resource_dict(data)
    unpacked_dict = serialization.JsonSerializer().deserialize(
      json_packed.contents)
    return resource_registry.from_resource_dict(unpacked_dict)

  @classmethod
  def pack(cls, resource: interfaces.ResourceBase) -> references.PackedResource:
    """Packs the resource."""
    ref = cls.from_resource(resource)
    return ref, references.PackedResource(
      JsonPackedResource(serialization.JsonSerializer().serialize(
        resource.to_resource_dict())).to_resource_dict(),
      ref_dict=ref.to_resource_dict(),
      links={})


class RefTest(unittest.TestCase):
  """A test for refs."""

  def test_verify_pack_unpack(self):
    """Test that packing and unpacking works."""
    resource = SimpleResource(x=1, y=2.0)
    ref, packed = enact.Ref.pack(resource)
    packed.ref().verify(packed.data)
    self.assertEqual(ref, packed.ref())
    unpacked = packed.ref().unpack(packed)
    self.assertEqual(unpacked, resource)

  def test_custom_ref_type(self):
    """Tests custom ref types works."""
    resource = SimpleResource(x=1, y=2.0)
    ref, packed = JsonPackedRef.pack(resource)
    self.assertEqual(ref, packed.ref())
    unpacked = packed.ref().unpack(packed)
    packed.ref().verify(packed.data)
    self.assertEqual(unpacked, resource)

  def test_id_from_id(self):
    """Test that references can be cast to and from ids."""
    ref: enact.Ref = enact.Ref('fake_digest')
    ref2 = ref.from_id(ref.id)
    self.assertEqual(ref, ref2)

  def test_ref_non_string_digest(self):
    """Test that references cannot be constructed from a non-string digest."""
    with self.assertRaisesRegex(AssertionError, 'string digest'):
      _ = enact.Ref(1234)  # type: ignore

  def test_ref_to_wrapped_type(self):
    """Tests references to wrapped python types."""
    value = 1
    ref, packed = enact.Ref.pack(resource_registry.wrap(value))
    self.assertIsInstance(
      resource_registry.from_resource_dict(packed.data),
      resource_registry.IntWrapper)
    self.assertIsInstance(ref, enact.Ref)
    self.assertEqual(ref, packed.ref())
    packed.ref().verify(packed.data)
    self.assertEqual(packed.ref().unpack(packed), 1)


class StoreTest(unittest.TestCase):
  """A test for stores."""

  def test_commit_has_get(self):
    """Create a store and test that commit, has, and get work."""
    store = enact.Store()
    resource = SimpleResource(x=1, y=2.0)
    ref = store.commit(resource)
    self.assertTrue(store.has(ref))
    self.assertEqual(store.checkout(ref), resource)
    self.assertNotEqual(id(store.checkout(ref)), id(resource))

  def test_store_provided_backend(self):
    """Tests that constructing stores with a provided backend works."""
    b1 = enact.InMemoryBackend()
    self.assertEqual(
      enact.Store(b1)._backend,  # pylint: disable=protected-access
      b1)

    with tempfile.TemporaryDirectory() as tmpdir:
      b2 = enact.FileBackend(tmpdir)
      self.assertEqual(
        enact.Store(b2)._backend,  # pylint: disable=protected-access
        b2)

  def test_custom_ref(self):
    """Test stores with custom ref types."""
    store = enact.Store(ref_type=JsonPackedRef)
    resource = SimpleResource(x=1, y=2.0)
    ref = store.commit(resource)
    self.assertIsInstance(ref, JsonPackedRef)
    self.assertTrue(store.has(ref))
    self.assertEqual(store.checkout(ref), resource)
    self.assertNotEqual(id(store.checkout(ref)), id(resource))

  def test_wrapped_ref(self):
    store = enact.Store()
    for val in [None, 0, 0.0, 'str', True, bytes([1, 2, 3]),
                [1, 2, 3], {'a': 1}]:
      with self.subTest(val):
        ref = store.commit(val)
        self.assertEqual(store.checkout(ref), val)

  def test_caching(self):
    """Test that caching works correctly."""
    store = enact.Store()
    resource = SimpleResource(x=1, y=2.0)
    ref = store.commit(resource)

    # checkout() works because cached reference is correct.
    ref.checkout().x = 10

    # checkout() fails because cached reference is incorrect.
    with self.assertRaises(contexts.NoActiveContext):
      ref.checkout()

    with store:
      # Refetches the correct resource from the store.
      self.assertEqual(ref.checkout().x, 1)

  def test_modify(self):
    """Tests the modify context."""
    store = enact.Store()
    resource = SimpleResource(x=1, y=2.0)
    ref = store.commit(resource)
    old_digest = ref.digest

    with store:
      with ref.modify() as resource:
        resource.x = 10
      self.assertEqual(ref.checkout().x, 10)
    self.assertNotEqual(ref.digest, old_digest)

  def test_modify_wrapped(self):
    """Tests that modifying wrapped resources works."""
    store = enact.Store()
    with store:
      ref = store.commit([0, 1, 2])
      old_ref = enact.deepcopy(ref)
      with ref.modify() as elems:
        elems.append(3)
      self.assertEqual(ref.checkout(), [0, 1, 2, 3])
      self.assertNotEqual(ref, old_ref)

  def test_pack_none(self):
    """Tests packing none."""
    store = enact.Store()
    value = None
    ref = store.commit(value)
    _, packed = ref.pack(resource_registry.wrap(value))
    ref.verify(packed.data)

  def test_packed_links(self):
    """Tests links are properly tracked during packing."""
    store = enact.Store()
    with store:
      x = enact.commit(1)
      y = enact.commit(2.0)
      r1 = SimpleResource(x, [{'test': (y, y)}])  # type: ignore
      _, packed = enact.Ref.pack(r1)
      got = packed.links
      want = {x.id, y.id}
      self.assertEqual(got, want)

  def test_file_backend(self):
    """Tests the file backend."""
    with tempfile.TemporaryDirectory() as tmpdir:
      store = enact.Store(backend=enact.FileBackend(tmpdir))
      resource = SimpleResource(x=1, y=2.0)
      ref = store.commit(resource)
      self.assertTrue(store.has(ref))
      self.assertEqual(store.checkout(ref), resource)

  def test_commit_cyclic_fails(self):
    """Tests that commits with cylic resource graphs fail."""
    with tempfile.TemporaryDirectory() as tmpdir:
      r1 = SimpleResource(x=1, y=2.0)
      r2 = SimpleResource(x=1, y=2.0)
      r1.x = r2  # type: ignore
      r2.x = r1  # type: ignore

      with enact.Store(backend=enact.FileBackend(tmpdir)):
        with self.assertRaises(acyclic.CycleDetected):
          enact.commit(r1)

  def test_ref_distribution_key(self):
    """Makes sure that refs have a distribution key."""
    self.assertEqual(
      enact.Ref.type_key().distribution_key,
      types.TypeKey(version.DIST_NAME, version.__version__))

  def test_backend_get_types(self):
    """Tests that getting types work."""
    backend = enact.InMemoryBackend()
    with enact.Store(backend):
      r1 = enact.commit(
        SimpleResource(
          1,
          SimpleResource(
            2,
            SimpleResource(0, {1, 2}))))  # set  # type: ignore
      r2 = enact.commit([(1,2)])  # tuple in list
      r3 = enact.Ref.from_id('{"digest": "fake_id"}')
      # pylint: disable=unbalanced-tuple-unpacking
      t1, t2, t3 = backend.get_type_keys((r1.id, r2.id, r3.id))
      self.assertIsNone(t3)

      expected_t1 = {
        enact.Ref.type_key(),
        SimpleResource.type_key(),
        type_wrappers.SetWrapper.type_key(),
      }
      self.assertEqual(t1, expected_t1)

      expected_t2 = {
        enact.Ref.type_key(),
        resource_registry.ListWrapper.type_key(),
        type_wrappers.TupleWrapper.type_key()
      }
      self.assertEqual(t2, expected_t2)

  def test_backend_get_dependency_graph(self):
    """Tests that getting dependency graphs works."""
    backend = enact.InMemoryBackend()

    with enact.Store(backend):
      r1 = enact.commit(5)
      r2 = enact.commit([r1])
      r3 = enact.commit([r1])
      r4 = enact.commit([r2, r3, r1])
      fake_ref = enact.Ref.from_id('{"digest": "fake_id"}')

      graph = backend.get_dependency_graph(
        (r1.id, r2.id, r3.id, r4.id, fake_ref.id))
      expected_graph = {
        r4.id: {r2.id, r3.id, r1.id},
        r3.id: {r1.id},
        r2.id: {r1.id},
        r1.id: set(),
        fake_ref.id: None}

      self.assertEqual(graph, expected_graph)

  def test_backend_get_dependency_graph_depth(self):
    """Tests that getting dependency graphs up to a certain depth works."""
    backend = enact.InMemoryBackend()

    with enact.Store(backend):
      refs = [enact.commit(0)]
      for _ in range(20):
        refs.append(enact.commit(refs[-1]))
      graph = backend.get_dependency_graph((refs[-1].id,), max_depth=10)
      self.assertEqual(set(graph.keys()), {ref.id for ref in refs[-11:]})

  def test_get_transitive_type_requirements(self):
    """Tests that getting transitive type requirements works."""
    with enact.Store() as store:
      r1 = enact.commit(5)
      r2 = enact.commit([r1])
      r3 = enact.commit(SimpleResource(1, 2.0))
      r4 = enact.commit({r2, r3})

      expected_type_requirements = {
        enact.Ref.type_key(),
        resource_registry.IntWrapper.type_key(),
        resource_registry.ListWrapper.type_key(),
        SimpleResource.type_key(),
        type_wrappers.SetWrapper.type_key(),
      }

      type_requirements = store.get_transitive_type_requirements(r4)
      self.assertEqual(type_requirements, expected_type_requirements)

  def test_get_transitive_type_requirements_fails(self):
    """Tests that the transitive type fails if a reference is not present."""
    with enact.Store() as store:
      fake_ref = enact.Ref.from_id('{"digest": "fake_id"}')
      r = enact.commit(fake_ref)
      with self.assertRaises(references.NotFound):
        store.get_transitive_type_requirements(r)

  def test_get_distribution_requirements(self):
    """Tests that getting distribution requirements works."""
    with enact.Store() as store:
      r1 = enact.commit(5)
      r2 = enact.commit([r1])
      r3 = enact.commit(SimpleResource(1, 2.0))
      r4 = enact.commit({r2, r3})

      # We don't have a distribution key for SimpleResource.
      with self.assertRaises(references.DistributionKeyError):
        store.get_distribution_requirements(r4)

      expected_dist_requirements = {
        types.DistributionKey(version.DIST_NAME, version.__version__),
      }

      dist_requirements = store.get_distribution_requirements(
        r4, expect_distribution_key=False)

      self.assertEqual(dist_requirements, expected_dist_requirements)

  def test_type_registration(self):
    with tempfile.TemporaryDirectory() as tmp_dir:
      backends: List[references.StorageBackend] = [
        enact.FileBackend(tmp_dir), enact.InMemoryBackend()]

      for backend in backends:
        with self.subTest(backend=type(backend).__name__):
          with enact.Store(backend):
            _ = enact.commit(SimpleResource(1, 2.0))
            attributes = backend.get_type(SimpleResource.type_key())
            self.assertEqual(
              attributes,
              {
                'x': types.Int(),
                'y': types.Float()
              }
            )



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Invocation generators for handling input requests."""

from typing import Any, Callable, Dict, Generic, Optional, Sequence, TypeVar, cast

from enact import function_wrappers, invocations, resource_registry
from enact import references

# Input value type.
I_contra = TypeVar('I_contra', contravariant=True)
# Output value type.
O_co = TypeVar('O_co', covariant=True)


class InvocationGenerator(Generic[I_contra, O_co]):
  """A generator that yields InputRequests from an invocation."""

  @classmethod
  def from_callable(
      cls,
      fun: Callable,
      args: Sequence=(),
      kwargs: Optional[Dict[str, Any]]=None) -> 'InvocationGenerator':
    """Create an InvocationGenerator from a callable.

    Args:
      fun: The callable to generate an invocation from.
      args: The arguments to the callable.
      kwargs: The keyword arguments to the callable.
    """
    kwargs = kwargs or {}
    if isinstance(fun, invocations.AsyncInvokableBase):
      raise TypeError(
        'AsyncInvokables are not currently supported by InvocationGenerator.')
    if isinstance(fun, invocations.InvokableBase):
      if not len(args) >= 1: raise ValueError(
        'Cannot specify more than one argument for an invokable.')
      if kwargs:
        raise ValueError('May not use keyword arguments with an invokable.')
      arg = args[0] if args else None
      return InvocationGenerator(
        invokable=fun,
        input_ref=references.commit(arg))
    if not function_wrappers.get_wrapper_type(fun):
      raise ValueError(
        'Cannot use invocation generator with an unregistered function.')
    invocation = function_wrappers.invoke(fun, args, kwargs)
    return cls.from_invocation(invocation)

  @classmethod
  def from_invocation(
      cls,
      invocation: invocations.Invocation[I_contra, O_co]) -> (
        'InvocationGenerator[I_contra, O_co]'):
    return InvocationGenerator(from_invocation=invocation)

  def __init__(
      self,
      invokable: Optional[invocations.InvokableBase[I_contra, O_co]]=None,
      input_ref: Optional[references.Ref[I_contra]]=None,
      from_invocation: Optional[invocations.Invocation[I_contra, O_co]]=None):
    """Initializes an interactive invocation."""
    if invokable and not isinstance(invokable, invocations.InvokableBase):
      if isinstance(invokable, invocations.AsyncInvokableBase):
        raise TypeError(
          'AsyncInvokables cannot be used with InvocationGenerator.')
      raise TypeError(
        'Invokable must be a child of Syncinvocations.InvokableBase.')
    if from_invocation is not None and invokable is not None:
      raise ValueError(
        'Cannot specify both an invokable and an invocation.')
    if from_invocation is not None and input_ref is not None:
      raise ValueError(
        'Cannot specify both an input and an invocation.')
    self._invocation: Optional[invocations.Invocation[I_contra, O_co]] = None
    self._from_invocation = from_invocation
    self._invokable = invokable
    self._input = input_ref
    self._request_input: Any = None

  @property
  def complete(self) -> bool:
    """Whether the invocation is complete."""
    if not self._invocation:
      return False
    if self._invocation.successful():
      return True
    if not self._invocation.response().raised:
      return False
    return not isinstance(self._invocation.get_raised(),
                          invocations.InputRequest)

  @property
  def invocation(self) -> invocations.Invocation[I_contra, O_co]:
    """The invocation."""
    if not self._invocation:
      raise invocations.InvocationError(
        'Invocation not started, please call next.')
    return self._invocation

  @property
  def input_request(self) -> invocations.InputRequest:
    """The current input request."""
    if self.complete:
      raise invocations.InvocationError('Invocation is complete.')
    input_request = self.invocation.get_raised()
    assert isinstance(input_request, invocations.InputRequest)
    return input_request

  def set_input(self, value: Any):
    """Set input for the next call to __next__.

    This allows using the generator in iterator-style, e.g.,

      for input_request in invocation_generator:
        input_request.set_input(...)

    Which can be more convenient than using send():

      input_request = next(invocation_generator)
      while True:
        try:
          input_request.send(...)
        except StopIteration:
          break

    Args:
      resource: The resource to set as input.
    """
    self._request_input = resource_registry.deepcopy(value)

  def __iter__(self) -> 'InvocationGenerator[I_contra, O_co]':
    """Returns the generator."""
    return self

  def __next__(self) -> invocations.InputRequest:
    """Continues the invocation until the next input request or completion."""
    if self.complete:
      raise StopIteration()

    if not self._invocation:
      if self._from_invocation:
        self._invocation = self._from_invocation.replay()
      else:
        assert self._invokable
        if not self._input:
          if self._invokable.get_input_type() != type(None):
            raise invocations.InvocationError(
              'Invokable has non-None input type. '
              'Please provide an explicit input reference on generator '
              'construction')
          self._input = cast(references.Ref[I_contra],
                             references.commit(None))
        self._invocation = self._invokable.invoke(self._input)
      if self.complete:
        raise StopIteration()
      return self.input_request
    else:
      if self._request_input is None:
        if not self.input_request.requested_type is type(None):
          raise invocations.InvocationError(
            'Invocation requests non-None input. Please use \'send(...)\' '
            'instead or set the input using \'set_input(...)\'.')
      self._invocation = self.input_request.continue_invocation(
        self._invocation, self._request_input)
      self._request_input = None
      if self.complete:
        raise StopIteration()
      return self.input_request

  def send(self, value) -> invocations.InputRequest:
    if not self._invocation:
      if value is not None:
        raise TypeError(
          'Can\'t send non-None value to a just-started generator.')
      return next(self)
    if self.complete:
      raise StopIteration()
    if not isinstance(value, self.input_request.requested_type):
      raise invocations.InvokableTypeError(
        f'Input type {type(value)} does not match requested type: '
        f'{self.input_request.requested_type}.')
    self._invocation = self.input_request.continue_invocation(
      self.invocation, value)
    if self.complete:
      raise StopIteration()
    return self.input_request



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions and classes."""
from typing import Any, Dict, Iterable, Optional

from enact import acyclic
from enact import interfaces
from enact import resource_digests
from enact import resource_registry
from enact import types

# pylint: disable=invalid-name
class cached_property(property):
  """A property that caches its result based on the digest of the resource.

  NOTE: This is effectively an LRU cache with a maxsize of 1, so the cache
  is cleared when the digest of the resource changes.
  """
  def __init__(self, user_function):
    self._user_function = user_function
    self._cache: Dict[str, Any] = {}
    super().__init__()

  def __get__(self, instance: Any, unused_owner: Optional[type] = None):
    """Cached property based on the digest of the resource."""
    digest = resource_digests.resource_digest(instance)
    result = self._cache.get(digest, None)
    if result is None:
      result = self._user_function(instance)
      self._cache = {digest: result}
    return result


def walk_resource_dict(
    value: interfaces.ResourceDictValue,
    include_self: bool = True) -> Iterable[interfaces.ResourceDict]:
  """Recursively yields all ResourceDict occurrences in the dictionary tree.

  Args:
    value: The value to walk.
    include_self: Whether to also yield 'value' if it is a resource dict.

  Yields:
    All instances of ResourceDict in the subtree defined by 'value' in
    depth-first order.
  """
  with acyclic.AcyclicContext(value):
    if isinstance(value, interfaces.ResourceDict) and include_self:
      yield value
    if isinstance(value, dict):  # Both normal dicts and resource dicts.
      for v in value.values():
        yield from walk_resource_dict(v, include_self=True)
    elif isinstance(value, list):
      for v in value:
        yield from walk_resource_dict(v, include_self=True)


def walk_resource(
    value: Any,
    include_self: bool = True) -> Iterable[interfaces.ResourceBase]:
  """Recursively yields all sub-resources in depth-first order.

  Will auto-wrap non-primitive python values.

  Args:
    value: The value to walk.
    include_self: Whether to also yield 'value' if it is a resource dict.

  Yields:
    All instances of ResourceBase in the subtree defined by 'value' in
    depth-first order.
  """
  with acyclic.AcyclicContext(value):
    if isinstance(value, dict):
      for v in value.values():
        yield from walk_resource(v, include_self=True)
    elif isinstance(value, list):
      for v in value:
        yield from walk_resource(v, include_self=True)
    elif isinstance(value, types.PRIMITIVES):
      pass
    elif isinstance(value, type):
      pass
    else:
      value = resource_registry.wrap(value)
      if include_self:
        yield value
      assert isinstance(value, interfaces.ResourceBase)
      for v in value.field_values():
        yield from walk_resource(v, include_self=True)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pretty-printing for resources and references."""

import dataclasses
from typing import Any, Callable, List, Optional, Set, Tuple, Type

from enact import function_wrappers
from enact import interfaces
from enact import invocations
from enact import references
from enact import resource_registry
from enact import types


@dataclasses.dataclass
class PPValue:
  """Marks an indented group."""
  value: str
  contents: List['PPValue'] = dataclasses.field(default_factory=list)
  open: str = ''
  close: str = ''


class PPrinter:
  """A pretty-printer."""

  def __init__(
      self, offset: int=2,
      max_ref_depth: Optional[int]=None,
      skip_repeated_refs: bool=False):
    """Initializes the pretty-printer."""
    self._formatters: List[Tuple[
        Type, Callable[[interfaces.FieldValue, int], PPValue]]] = [
      (str, self.from_primitive),
      (bytes, self.from_primitive),
      (references.Ref, self.from_ref),
      (interfaces.ResourceBase, self.from_resource),
      (dict, self.from_dict),
      (list, self.from_list),
      (type, self.from_type),
    ]

    self.offset = offset
    self.max_ref_depth = max_ref_depth
    self.skip_repeated_refs = skip_repeated_refs
    self._seen_refs: Set[str] = set()

  def from_type(self, v: interfaces.FieldValue, unused_depth: int) -> PPValue:
    assert isinstance(v, type)
    assert issubclass(v, interfaces.ResourceBase), (
      'Only types that are subclasses of ResourceBase are supported.')
    return PPValue(v.__name__, [])

  def primitive_to_str(self, v: types.Primitives) -> str:
    """Converts a primitive to a string."""
    if isinstance(v, bytes):
      return f'<{len(v)} bytes>'
    if isinstance(v, str):
      return repr(v)
    return str(v)

  def from_list(self, v: interfaces.FieldValue, depth: int) -> PPValue:
    assert isinstance(v, list)
    return PPValue(
      '', [self.pvalue(item, depth + 1) for item in v], '[', ']')

  def from_dict(self, v: interfaces.FieldValue, depth: int) -> PPValue:
    assert isinstance(v, dict)
    return PPValue(
      '',
      [PPValue(
        f'"{key}":',
        [self.pvalue(item, depth + 1)]) for key, item in v.items()],
      '{', '}')

  def from_resource(self, v: interfaces.FieldValue, depth: int) -> PPValue:
    assert isinstance(v, interfaces.ResourceBase)
    if all(isinstance(field_value, types.PRIMITIVES)
           for field_value in v.field_values()):
      str_values = ', '.join([
        f'{field_name}={self.primitive_to_str(field_value)}'  # type: ignore
        for field_name, field_value in v.field_items()])
      return PPValue(f'{type(v).__name__}({str_values})', [])
    result = PPValue(
      value=type(v).__name__,
      contents=[
        PPValue(field_name, [self.pvalue(field_value, depth + 1)], open=':')
        for field_name, field_value in v.field_items()],
      open=':',
      close='')
    return result

  def from_ref(self, v: interfaces.FieldValue, depth: int) -> PPValue:
    assert isinstance(v, references.Ref)
    resource: interfaces.ResourceBase = v()
    if ((self.max_ref_depth and depth > self.max_ref_depth) or
        (self.skip_repeated_refs and v.digest in self._seen_refs)):
      return PPValue(f'-> {type(resource).__name__}#{v.digest[0:6]}', [])
    result = self.pvalue(resource, depth + 1)
    result.value = f'-> {result.value}#{v.digest[0:6]}'
    self._seen_refs.add(v.digest)
    return result

  def from_primitive(
      self, v: interfaces.FieldValue, unused_depth: int) -> PPValue:
    assert isinstance(v, types.PRIMITIVES)
    return PPValue(self.primitive_to_str(v), [])

  def pvalue(
      self, v: Any, depth: int=0) -> PPValue:
    """Produces a nested string value for pretty-printing."""
    field_value = resource_registry.to_field_value(v)
    for t, formatter in self._formatters:
      if isinstance(field_value, t):
        return formatter(field_value, depth)
    return self.from_primitive(field_value, depth)

  def _merge(self, v: PPValue) -> List[Tuple[int, str]]:
    """Recursively merge pvalues."""
    result: List[Tuple[int, str]] = [(0, v.value + v.open)]
    for c in v.contents:
      result += [
        (d + 1, merged_c) for d, merged_c in self._merge(c)]
    result[-1] = (result[-1][0], result[-1][1] + v.close)
    if len(result) == 2:
      result = [(result[0][0], result[0][1] + ' ' + result[1][1])]
    return result

  def register(
      self,
      t: Type,
      formatter: Callable[[interfaces.FieldValue, int], PPValue]) -> None:
    """Register a new formatter."""
    self._formatters.insert(0, (t, formatter))

  def pformat(self, v: Any) -> str:
    """Return a pretty string for an enact value."""
    self._seen_refs.clear()
    pvalue = self._merge(self.pvalue(v))
    return '\n'.join(
      ' ' * self.offset * d + s for d, s in pvalue)

  def pprint(self, v: Any) -> None:
    """Pretty-print an enact value."""
    print(self.pformat(v))


def pprint(value: Any,
           max_ref_depth: Optional[int]=None,
           skip_repeated_refs: bool=False) -> None:
  """Pretty-prints a resource, optionally resolving references."""
  PPrinter(max_ref_depth=max_ref_depth,
           skip_repeated_refs=skip_repeated_refs).pprint(value)


def pformat(value: Any,
            max_ref_depth: Optional[int]=None,
            skip_repeated_refs: bool=False) -> str:
  """Pretty-prints a resource, optionally resolving references."""
  return PPrinter(
    max_ref_depth=max_ref_depth,
    skip_repeated_refs=skip_repeated_refs).pformat(value)


def invocation_summary(invocation: invocations.Invocation, indent: int=0):
  """Return a reader-friendly invocation summary."""
  invokable = f'->{invocation.request().invokable()}'
  input_value = invocation.get_input()
  if isinstance(input_value, function_wrappers.CallArgs):
    args, kwargs = input_value.to_python_args()
    str_args = (
      [str(a) for a in args] +
      [f'{k}={str(v)}' for k, v in kwargs.items()])
    all_args = ', '.join(str_args)
  else:
    all_args = str(input_value)
  if invocation.successful():
    output = f'= {invocation.get_output()}'
  elif invocation.response().raised:
    output = f'raised {invocation.get_raised()}'
  else:
    output = 'incomplete'

  prefix = '  ' * indent
  lines = [
    f'{prefix}{invokable}({all_args}) {output}',
    *[f'{invocation_summary(child, indent + 1)}'
      for child in invocation.get_children()]]
  return '\n'.join(lines)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for invocation generators."""

import dataclasses
import tempfile
import unittest

import enact


@enact.typed_invokable(int, str)
@dataclasses.dataclass
class IntToStr(enact.Invokable):
  """An invokable that converts an int to a string."""
  salt: str = ''

  def call(self, value: int) -> str:
    return str(value) + self.salt


class InvocationGeneratorTest(unittest.TestCase):
  """Tests invocation_generator"""

  def setUp(self):
    # pylint: disable=consider-using-with
    self.dir = tempfile.TemporaryDirectory()
    self.backend = enact.FileBackend(self.dir.name)
    self.store = enact.Store(self.backend)

  def tearDown(self):
    self.dir.cleanup()

  def test_send_without_next(self):
    """Tests that send without next fails on invokable generator."""
    with self.store:
      inv_gen = enact.InvocationGenerator(
        IntToStr(), enact.commit(3))
      with self.assertRaisesRegex(TypeError, '.*non-None.*'):
        inv_gen.send(3)

  def test_send_none_without_next(self):
    """Tests that send None without next works."""
    with self.store:
      inv_gen = enact.InvocationGenerator(
        IntToStr(), enact.commit(3))
      with self.assertRaises(StopIteration):
        inv_gen.send(None)

  def test_send_flow(self):
    """Test an invokable generator in a send-based flow."""
    @enact.typed_invokable(type(None), int)
    class SumUserRequests(enact.Invokable):
      def call(self):
        return (
          enact.request_input(int) +
          enact.request_input(int) +
          enact.request_input(int))

    with self.store:
      inv_gen = enact.InvocationGenerator(
        SumUserRequests(), enact.commit(None))
      input_request = next(inv_gen)
      for i in range(5):
        assert isinstance(input_request, enact.InputRequest)
        self.assertFalse(inv_gen.complete)
        try:
          input_request = inv_gen.send(i)
        except StopIteration:
          break
      self.assertEqual(
        inv_gen.invocation.get_output(), 0 + 1 + 2)

  def test_set_input_flow(self):
    """Test an invokable generator in a send-based flow."""
    @enact.typed_invokable(type(None), int)
    class SumUserRequests(enact.Invokable):
      def call(self):
        return (
          enact.request_input(int) +
          enact.request_input(int) +
          enact.request_input(int))

    with self.store:
      inv_gen = enact.InvocationGenerator(
        SumUserRequests(), enact.commit(None))
      for i, _ in enumerate(inv_gen):
        self.assertFalse(inv_gen.complete)
        inv_gen.set_input(i)
      self.assertEqual(
        inv_gen.invocation.get_output(), 0 + 1 + 2)

  def test_wrapped_functions(self):
    """Tests invocation generator for wrapped functions."""
    @enact.register
    def request_plus_one() -> int:
      return enact.request_input(int) + 1

    @enact.register
    def sum_requests() -> int:
      return request_plus_one() + request_plus_one()

    with self.store:
      inv_gen: enact.InvocationGenerator = (
          enact.InvocationGenerator.from_callable(sum_requests))
      for i, _ in enumerate(inv_gen):
        self.assertFalse(inv_gen.complete)
        inv_gen.set_input(i)
      invocation = inv_gen.invocation
      self.assertEqual(invocation.get_output(), 1 + 2)
      self.assertEqual(invocation.get_child(0).get_output(), 1)
      self.assertEqual(invocation.get_child(1).get_output(), 2)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Thread-safe context management."""

import contextlib
import contextvars
from typing import Callable, Dict, Optional, Type, TypeVar, cast


_context_vars: Dict[
  Type['_ContextBase'],
  contextvars.ContextVar[Optional['_ContextBase']]] = {}


class ContextError(Exception):
  """Error raised when there is a problem with the context."""

class ContextTypeError(ContextError):
  """Error raised when the context has an unexpected type"""


class NoActiveContext(ContextError):
  """Raised when there is no active context."""

ContextBaseT = TypeVar('ContextBaseT', bound='_ContextBase')
ContextSuperT = TypeVar('ContextSuperT', bound='_ContextBase')
ContextT = TypeVar('ContextT', bound='Context')
AsyncContextT = TypeVar('AsyncContextT', bound='AsyncContext')


class _ContextBase:
  """A thread-aware context superclass."""

  def __init__(self: ContextBaseT):
    """Creates a new context."""
    self._token: Optional[contextvars.Token[Optional[ContextBaseT]]] = None

  @classmethod
  def _get_context_var(cls: Type[ContextBaseT]) -> (
      contextvars.ContextVar[Optional[ContextBaseT]]):
    """Returns the context var for this type."""
    try:
      return cast(
        contextvars.ContextVar[Optional[ContextBaseT]], _context_vars[cls])
    except KeyError as key_error:
      raise ContextError(
        f'Context {cls} not registered. A context class must be registered '
        f'with the "@register" decorator.') from key_error

  @classmethod
  def permissive_initialization(cls: Type[ContextBaseT]) -> bool:
    """Returns whether the class has permissive initialization.

    Contexts use contextvars for tracking whether an execution is in-context or
    out-of-context. Contextvars have special handling in threads and async
    executions, e.g., they are thread-local and hence starting a new thread will
    erase the context stack. The Context class can detect when this happens,
    since it will store information in the main thread during registration. This
    variable controls how it handles such situations:

    If permissive_initialization is set to True, the context will automatically
    initialize to an empty stack when it detects that it is being run in an
    uninitialized context. If set to False, then the class will trigger an error
    and ask the developer to specify whether they want to copy context or start
    a fresh calling context. This may be done using the 'with_current_contexts'
    or with_new_contexts function decorators.

    Overriding this function to return False is useful in situation where
    accidentally operating in a fresh context is a problem, e.g., when
    implementing contexts that deal with encryption.
    """
    return True

  @classmethod
  @contextlib.contextmanager
  def top_level(cls: Type[ContextBaseT]):
    """Returns a context manager to execute code in a top-level context."""
    context_var = cls._get_context_var()
    token = context_var.set(None)
    try:
      yield
    finally:
      context_var.reset(token)

  @classmethod
  def get_current(cls: Type[ContextBaseT]) -> Optional[ContextBaseT]:
    """Returns the current context of this type or None."""
    context_var = cls._get_context_var()
    try:
      current_context = context_var.get()
    except LookupError as lookup_error:
      if cls.permissive_initialization():
        context_var.set(None)
        return None
      raise ContextError(
        f'Context {cls} not initialized. If running inside a thread, make sure '
        f'to annotate the thread function with either the '
        f'"@with_current_contexts" or "@with_new_contexts" decorator.'
        ) from lookup_error
    if current_context is not None and not isinstance(current_context, cls):
      raise ContextTypeError(
        f'Current context {current_context} is not of type {cls}.')
    return current_context

  @classmethod
  def current(cls: Type[ContextBaseT]) -> ContextBaseT:
    """Returns the current context of this type or raises an error."""
    context = cls.get_current()
    if context is None:
      raise NoActiveContext(f'No context of type {cls.__qualname__} is active.')
    return context


class Context(_ContextBase):
  """A thread-aware context superclass."""

  def __enter__(self: ContextT) -> ContextT:
    """Enters the context."""
    context_var = self._get_context_var()
    try:
      self.get_current()  # Raise an error if the context is not initialized.
    except ContextTypeError:
      pass   #  We don't care here if the context has the wrong type.
    self.enter()
    self._token = context_var.set(self)
    return self

  def __exit__(self: ContextT, exc_type, exc_value, traceback):
    """Exits the context."""
    context_var = self._get_context_var()
    assert self._token
    context_var.reset(self._token)
    self.exit()
    self._token = None

  def enter(self):
    """Overridable on context entry."""

  def exit(self):
    """Overridable on context exit."""


class AsyncContext(_ContextBase):
  """A thread-aware async context superclass."""

  async def __aenter__(self: AsyncContextT) -> AsyncContextT:
    """Enters the context."""
    context_var = self._get_context_var()
    try:
      self.get_current()  # Raise an error if the context is not initialized.
    except ContextTypeError:
      pass   #  We don't care here if the context has the wrong type.
    await self.aenter()
    self._token = context_var.set(self)
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    """Exits the context."""
    context_var = self._get_context_var()
    assert self._token
    context_var.reset(self._token)  # type: ignore
    await self.aexit()
    self._token = None

  async def aenter(self):
    """Overridable on context entry."""

  async def aexit(self):
    """Overridable on context exit."""



def register(cls: Type[ContextBaseT]) -> Type[ContextBaseT]:
  """Registers a context class."""
  assert cls not in _context_vars, (
    f'Context class already registered: {cls}')
  ctx_var: contextvars.ContextVar[Optional[_ContextBase]] = (
    contextvars.ContextVar(cls.__qualname__))
  ctx_var.set(None)
  _context_vars[cls] = ctx_var
  return cls


def register_to_superclass(superclass: Type[ContextSuperT]) -> Callable[
    [Type[ContextBaseT]], Type[ContextBaseT]]:
  def _register(cls: Type[ContextBaseT]) -> Type[ContextBaseT]:
    """Registers a context class."""
    assert cls not in _context_vars, (
      f'Context class already registered: {cls}')
    assert superclass in _context_vars, (
      f'Superclass {superclass} not registered.')
    assert issubclass(cls, superclass), (
      f'Context class {cls} must be a subclass of {superclass}.')
    _context_vars[cls] = _context_vars[superclass]
    # MyPy can't handle intersection types, which is what would be required
    # here. We cast to Type[ContextT] to make it happy.
    return cast(Type[ContextBaseT], cls)
  return _register


R = TypeVar('R')


def _with_contexts(
    ctx: contextvars.Context,
    fun: Callable[..., R], *args, **kwargs) -> R:
  """Temporarily set _context_vars from ctx and run function."""
  # We don't use ctx.run here, since we don't want to modify how unmanaged
  # ContextVar objects are propagated.
  tokens: Dict[contextvars.ContextVar, contextvars.Token] = {}
  for context_var in _context_vars.values():
    assert context_var in ctx
    tokens[context_var] = context_var.set(ctx[context_var])
  try:
    return fun(*args, **kwargs)
  finally:
    for context_var, token in tokens.items():
      context_var.reset(token)


def with_current_contexts(fun: Callable) -> Callable:
  """Decorator that runs the function in all current contexts."""
  ctx = contextvars.copy_context()
  return lambda *args, **kwargs: _with_contexts(ctx, fun, *args, **kwargs)


def with_new_contexts(fun: Callable) -> Callable:
  """Decorator that runs the function with new contexts."""
  ctx = contextvars.Context()
  def _set_to_none():
    for context_var in _context_vars.values():
      context_var.set(None)
  ctx.run(_set_to_none)  # Set the managed contextvars to None.
  return lambda *args, **kwargs: _with_contexts(ctx, fun, *args, **kwargs)



"""Hashing functions for Enact resources."""
from typing import Any

from enact import digests
from enact import resource_registry


def resource_digest(value: Any) -> str:
  """Compute a hash digest of a resource."""
  return digests.digest(resource_registry.wrap(value))



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checks that the package version and python version match."""

import os
import unittest
import re

import enact

class TestVersion(unittest.TestCase):
  """Tests for the version."""

  def test_version(self):
    """Tests that the version matches."""
    try:
      pyproject_path = os.path.join(
          os.path.dirname(__file__), os.pardir, 'pyproject.toml')
      with open(pyproject_path, 'r', encoding='utf8') as f:
        pyproject_contents = f.read()
    except FileNotFoundError:
      self.skipTest('pyproject.toml not found')
    # Python 3.8 does not have tomllib, so we use a regex instead.
    pyproject_contents = re.search(r'version = "(.*)"', pyproject_contents)
    pyproject_version = pyproject_contents.group(1)
    self.assertEqual(enact.__version__, pyproject_version)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Top-level definitions."""

from enact.distribution_registry import register_distribution
from enact.distribution_registry import get_path_distribution_key
from enact.distribution_registry import get_distribution_key
from enact.types import DistributionKey
from enact.types import TypeKey
from enact.types import TypeDescriptor

from enact.version import __version__
from enact.function_wrappers import invoke
from enact.function_wrappers import invoke_async
from enact.resource_digests import resource_digest

from enact.interfaces import FieldValue
from enact.interfaces import ResourceDictValue
from enact.interfaces import FieldTypeError
from enact.interfaces import ResourceBase
from enact.interfaces import TypeWrapperBase
from enact.interfaces import ResourceDict

from enact.invocations import _InvokableBase
from enact.invocations import AsyncInvokableBase
from enact.invocations import AsyncInvokable
from enact.invocations import ExceptionResource
from enact.invocations import IncompleteSubinvocationError
from enact.invocations import InputChanged
from enact.invocations import InputRequest
from enact.invocations import InputRequestOutsideInvocation
from enact.invocations import Invokable
from enact.invocations import Invocation
from enact.invocations import ReplayContext
from enact.invocations import ReplayError
from enact.invocations import Request
from enact.invocations import RequestedTypeUndetermined
from enact.invocations import Response
from enact.invocations import RequestInput
from enact.invocations import request_input
from enact.invocations import typed_invokable
from enact.invocations import InvokableBase
from enact.invocations import NativeException

from enact.invocation_generators import InvocationGenerator

from enact.pretty_print import pformat
from enact.pretty_print import pprint
from enact.pretty_print import PPrinter
from enact.pretty_print import PPValue
from enact.pretty_print import invocation_summary

from enact.references import commit
from enact.references import checkout
from enact.references import FileBackend
from enact.references import InMemoryBackend
from enact.references import Ref
from enact.references import RefError
from enact.references import Store
from enact.references import InMemoryStore
from enact.references import FileStore

from enact.resources import Resource
from enact.resources import ImmutableResource
from enact.resources import TypeWrapper

from enact.registration import register

from enact.resource_registry import Registry
from enact.resource_registry import RegistryError
from enact.resource_registry import ResourceNotFound
from enact.resource_registry import wrap
from enact.resource_registry import unwrap
from enact.resource_registry import deepcopy
from enact.resource_registry import to_python_type
from enact.resource_registry import from_python_type

# Trigger registration of standard resource wrappers.
import enact.type_wrappers

from enact.serialization import Serializer
from enact.serialization import JsonSerializer
from enact.serialization import SerializationError
from enact.serialization import DeserializationError

from enact import contexts



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for digests."""

import unittest

from enact import digests, resource_registry

import random_value  # type: ignore


class DigestTest(unittest.TestCase):
  """Tests for digests."""

  def test_equal_digest_resource_and_dict(self):
    """Tests that resources and their dicts have the same digest."""
    for _ in range(100):
      resource = random_value.rand_resource()
      self.assertEqual(digests.digest(resource),
                       digests.digest(resource.to_resource_dict()))

  def test_none_digest(self):
    """Ensures that none has the same digest as dict or resource."""
    none = resource_registry.wrap(None)
    resource_digest = digests.digest(none)
    dict_digest = digests.digest(none.to_resource_dict())
    self.assertEqual(resource_digest, dict_digest)
